from collections.abc import Iterable
from typing import Union

import abjad


def leaves_are_tieable(
    leaves: Union[
        abjad.Selection,
        Iterable[abjad.Component],
        Iterable[abjad.LogicalTie],
    ],
    *,
    only_identical_pitches: bool = False,
) -> bool:
    r"""Returns a :obj:`bool` representing whether or not two or more input
    leaves have any identical pitch(es) and thus can be tied. Input argument
    can be a single |abjad.Selection| with multiple leaves, or an iterable with
    elements of type |abjad.Leaf| or child classes.

    Basic usage:
        When one or more pitches of a leave is also present in the next one,
        the function returns ``True``:

        >>> leaf1 = abjad.Note(r"c'4")
        >>> leaf2 = abjad.Note(r"c'4")
        >>> auxjad.get.leaves_are_tieable([leaf1, leaf2])
        True

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.get|. It can thus be used from either |auxjad.get|_ or
        |abjad.get| namespaces. Therefore, the two lines below are equivalent:

        >>> leaf1 = abjad.Note(r"c'4")
        >>> leaf2 = abjad.Note(r"c'4")
        >>> auxjad.get.leaves_are_tieable([leaf1, leaf2])
        True
        >>> abjad.get.leaves_are_tieable([leaf1, leaf2])
        True

    Durations:
        Durations do not affect the comparison.

        >>> leaf1 = abjad.Note(r"c'2.")
        >>> leaf2 = abjad.Note(r"c'16")
        >>> leaf3 = abjad.Note(r"f'''16")
        >>> auxjad.get.leaves_are_tieable([leaf1, leaf2])
        True
        >>> auxjad.get.leaves_are_tieable([leaf1, leaf3])
        False
        >>> auxjad.get.leaves_are_tieable([leaf2, leaf3])
        False

    Chords:
        Handles chords.

        >>> chord1 = abjad.Chord(r"<c' e' g'>4")
        >>> chord2 = abjad.Chord(r"<c' e' g'>16")
        >>> chord3 = abjad.Chord(r"<f''' fs'''>16")
        >>> auxjad.get.leaves_are_tieable([chord1, chord2])
        True
        >>> auxjad.get.leaves_are_tieable([chord1, chord3])
        False
        >>> auxjad.get.leaves_are_tieable([chord2, chord3])
        False

    ``only_identical_pitches``:
        By default, if any pitch in a leaf (be it either a note or chord) can
        be tied to the next leaf, the function returns ``True``:

        >>> chord1 = abjad.Chord(r"<c' e' g'>4")
        >>> chord2 = abjad.Chord(r"<c' e' g' bf'>4")
        >>> note = abjad.Note(r"c'4")
        >>> auxjad.get.leaves_are_tieable([chord1, chord2])
        True
        >>> auxjad.get.leaves_are_tieable([chord2, note])
        True

        Set the argument ``only_identical_pitches`` to ``True`` so that the
        function only returns ``True`` for identical pitch collections:

        >>> chord1 = abjad.Chord(r"<c' e' g'>4")
        >>> chord2 = abjad.Chord(r"<c' e' g' bf'>4")
        >>> note = abjad.Note(r"c'4")
        >>> auxjad.get.leaves_are_tieable([chord1, chord2],
        ...                               only_identical_pitches=True,
        ...                               )
        False
        >>> auxjad.get.leaves_are_tieable([chord2, note],
        ...                               only_identical_pitches=True,
        ...                               )
        False
        >>> chord3 = abjad.Chord(r"<c' e' g' bf'>4")
        >>> chord4 = abjad.Chord(r"<c' e' g' bf'>4")
        >>> auxjad.get.leaves_are_tieable([chord3, chord4],
        ...                               only_identical_pitches=True,
        ...                               )
        True

    Rests:
        If rests are input, the return value is ``False``. It also handles
        multi-measure rests.

        >>> leaf = abjad.Note(r"c'4")
        >>> rest = abjad.Rest(r"r4")
        >>> auxjad.get.leaves_are_tieable([leaf, rest])
        False

    Parentage:
        Leaves can also be part of |abjad.Container|'s or child classes.

        >>> container = abjad.Container(r"r4 <c' e'>4 <c' e'>2")
        >>> auxjad.get.leaves_are_tieable([container[1], container[2]])
        True

    Multiple leaves:
        It accepts more than two leaves and will return ``True`` if all leaves
        can be tied sequentially (i.e. leaf one to leaf two, leaf two to leaf
        three, etc.):

        >>> leaf1 = abjad.Note(r"c'4")
        >>> leaf2 = abjad.Note(r"c'4")
        >>> leaf3 = abjad.Note(r"c'2.")
        >>> auxjad.get.leaves_are_tieable([leaf1, leaf2, leaf3])
        True

    Logical ties:
        Accepts leaves as well as |abjad.LogicalTie|'s:

        >>> leaf = abjad.Note(r"c'2.")
        >>> staff = abjad.Staff(r"c'4 ~ c'16")
        >>> logical_tie = abjad.select(staff).logical_tie(0)
        >>> auxjad.get.leaves_are_tieable([leaf, logical_tie])
        True

    Selections:
        Accepts a single |abjad.Selection| as input and will return ``True`` if
        all leaves can be tied sequentially (i.e. leaf one to leaf two, leaf
        two to leaf three, etc.):

        >>> staff = abjad.Staff(r"c'2 c'4. c'8")
        >>> auxjad.get.leaves_are_tieable(staff[:])
        True
        >>> staff = abjad.Staff(r"c'2 c'4. d'8")
        >>> auxjad.get.leaves_are_tieable(staff[:])
        False
    """
    if not isinstance(leaves, (abjad.Selection, Iterable)):
        raise TypeError(
            "argument must be 'abjad.Selection' or an iterable of "
            "'abjad.Component's or 'abjad.LogicalTie's"
        )
    for leaf in leaves:
        if not isinstance(leaf, (abjad.Component, abjad.LogicalTie)):
            raise TypeError(
                "argument must be 'abjad.Selection' or an "
                "iterable of 'abjad.Component's or "
                "'abjad.LogicalTie's"
            )
    if not isinstance(only_identical_pitches, bool):
        raise TypeError("'only_identical_pitches' must be 'bool'")
    if len(leaves) < 2:
        raise ValueError("argument must contain two or more leaves")

    for index, leaf1 in enumerate(leaves[:-1]):
        for leaf2 in leaves[index + 1 :]:
            if isinstance(leaf1, abjad.LogicalTie):
                leaf1 = leaf1[0]
            elif isinstance(leaf2, abjad.LogicalTie):
                leaf2 = leaf2[0]
            elif isinstance(leaf1, abjad.Rest):
                return False
            elif isinstance(leaf1, abjad.MultimeasureRest):
                return False
            elif isinstance(leaf1, (abjad.Note, abjad.Chord)) and isinstance(
                leaf2, (abjad.Note, abjad.Chord)
            ):
                try:
                    pitches_1 = [leaf1.written_pitch]
                except AttributeError:
                    pitches_1 = [p for p in leaf1.written_pitches]
                pitches_1.sort()
                try:
                    pitches_2 = [leaf2.written_pitch]
                except AttributeError:
                    pitches_2 = [p for p in leaf2.written_pitches]
                pitches_2.sort()
                if only_identical_pitches and pitches_1 != pitches_2:
                    return False
                elif not any([p in pitches_2 for p in pitches_1]):
                    return False
            else:
                return False

            # leaf1_graces = abjad.get.before_grace_container(leaf1)
            # leaf2_graces = abjad.get.before_grace_container(leaf2)
            # if not isinstance(leaf1_graces, type(leaf2_graces)):
            #     return False
    return True



"""
select
======

Auxjad's selection functions. These can be accesses via either:

>>> abjad.select.auxjad_function(container)
>>> auxjad.select.auxjad_function(container)
"""

import abjad

from .logical_selections import logical_selections

### EXTENSION FUNCTIONS ###

abjad.select.logical_selections = logical_selections



from typing import Optional, Union

import abjad

from .. import mutate
from ..core.GeneticAlgorithm import GeneticAlgorithm


class GeneticAlgorithmMusicMaker:
    r"""Uses two :class:`auxjad.GeneticAlgorithm`'s, one for pitch and another
    for attack points, in order to create musical cells. At each call of
    :meth:`__call__`, it iterates the genetic algorithms by one generation, and
    returns an |abjad.Selection| created with the fittest pitch and attack
    point individuals.

    ..  note::

        Many of the properties of this class reflect the behaviour of
        properties of :class:`GeneticAlgorithm`. Some have the same name as
        those in :class:`GeneticAlgorithm`, including:

        * :attr:`population_size` and :attr:`GeneticAlgorithm.population_size`
        * :attr:`select_n_parents` and
          :attr:`GeneticAlgorithm.select_n_parents`
        * :attr:`keep_n_parents` and :attr:`GeneticAlgorithm.keep_n_parents`
        * :attr:`mutation_chance` and :attr:`GeneticAlgorithm.mutation_chance`
        * :attr:`mutation_index` and :attr:`GeneticAlgorithm.mutation_index`
        * :attr:`evaluation_index` and
          :attr:`GeneticAlgorithm.evaluation_index`
        * :attr:`generation_number` and
          :attr:`GeneticAlgorithm.generation_number`
        * :attr:`scores` and :attr:`GeneticAlgorithm.scores`
        * :attr:`fittest_individual_score` and
          :attr:`GeneticAlgorithm.fittest_individual_score`

        Properties related to pitch and attach points will also have
        equivalents:

        * :attr:`pitch_target` and :attr:`attack_point_target` are equivalent
          to :attr:`GeneticAlgorithm.target`
        * :attr:`pitch_genes` and :attr:`attack_point_genes` are equivalent to
          :attr:`GeneticAlgorithm.genes`
        * :attr:`pitch_initial_individual` and
          :attr:`attack_point_initial_individual` are equivalent to
          :attr:`GeneticAlgorithm.initial_individual`
        * :attr:`pitch_population` and :attr:`attack_point_population` are
          equivalent to :attr:`GeneticAlgorithm.population`
        * :attr:`fittest_pitch_individual` and
          :attr:`fittest_attack_point_individual` are equivalent to
          :attr:`GeneticAlgorithm.fittest_individual`

    Basic usage:
        At its basic, this class needs a target and a list of genes for both
        pitches and attack points. The evaluation function will compare all
        individuals in the population against this target when scoring them.

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ... )
        >>> repr(maker)
        pitches: ["c'", "d'", "e'", "f'"]
        attack_points: [0, 4, 8, 12]
        >>> len(maker)
        4
        >>> notes = maker.target_music
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-ahfUfls3cq.png

        Calling the instance will apply the genetic algorithm process and
        output an |abjad.Selection| with the fittest  individual in the
        population.

        >>> notes = maker()
        >>> maker.generation_number
        0
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                f'4
                ~
                f'8.
                e'16
                ~
                e'8
                a'8
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-mIrHl4wwHA.png

        Subsequent calls will create new generations of individuals, always
        outputting the fittest measure.

        >>> notes = maker()
        >>> maker.generation_number
        1
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                d'16
                e'8.
                ~
                e'4
                ~
                e'16
                g'8.
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-hDxTq3Y2Ek.png

    :meth:`output_n`:
        The method :meth:`output_n` can be used to output ``n`` iterations of
        the process. They are output as a single |abjad.Selection|:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                r4
                r16
                c'8.
                d'16
                d'8.
                f'4
                c'2
                d'16
                c''8.
                f'4
                r16
                c'8.
                d'4
                f'4.
                f'8
                c'4
                d'8.
                e'16
                ~
                e'4
                e'4
                c'4
                d'4
                ~
                d'16
                e'8.
                e'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-PrfaIjhEbL.png

    :attr:`pitch_genes`:
        While :attr:`attack_point_genes` must always take a :obj:`list` of
        :obj:`int`'s, :attr:`pitch_genes` can take a variety of object types.
        The implementation of this class uses |abjad.LeafMaker|, so pitches can
        take any objects accepted by that class. These include :obj:`int` and
        :obj:`str` for pitches, ``None`` for rests, :obj:`tuple` for chords,
        etc.

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", None, "e'", ("g'", "bf'")],
        ...     pitch_genes=[None,
        ...                  "c'",
        ...                  "d'",
        ...                  "e'",
        ...                  "f'",
        ...                  "g'",
        ...                  ("g'", "a'"),
        ...                  ("g'", "bf'"),
        ...                  ("g'", "c''"),
        ...                  ],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     population_size=50,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                <g' c''>4
                g'4
                ~
                g'16
                <g' c''>8.
                <g' c''>4
                c'2
                e'8.
                r16
                r4
                c'4
                r4
                r16
                e'8.
                <g' a'>4
                c'4
                r4
                e'8.
                <g' a'>16
                ~
                <g' a'>4
                c'4
                r4
                e'8.
                <g' a'>16
                ~
                <g' a'>4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-etoHzdnAIu.png

        Which is equivalent to:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=[0, None, 4, (7, 10)],
        ...     pitch_genes=[None, 0, 2, 4, 5, 7, (7, 9), (7, 10), (7, 12)],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     population_size=50,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                <g' c''>4
                g'4
                ~
                g'16
                <g' c''>8.
                <g' c''>4
                c'2
                e'8.
                r16
                r4
                c'4
                r4
                r16
                e'8.
                <g' a'>4
                c'4
                r4
                e'8.
                <g' a'>16
                ~
                <g' a'>4
                c'4
                r4
                e'8.
                <g' a'>16
                ~
                <g' a'>4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-AfDY1QYNhW.png

    :attr:`units_per_window` and :attr:`duration_unit`:
        By default, there are ``16`` attack points in a window, each lasting
        for ``abjad.Duration((1, 16))``. These can be changed using
        :attr:`units_per_window` and :attr:`duration_unit`:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 8, 16, 24],
        ...     attack_point_genes=list(range(32)),
        ...     duration_unit=abjad.Duration((1, 32)),
        ...     units_per_window=32,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8.
                e'16
                ~
                e'16.
                e'32
                ~
                e'8
                ~
                e'32
                b'4...
                c'8.
                d'16
                ~
                d'2
                f'4
                c'4
                d'8..
                e'32
                ~
                e'4
                c'4
                c'4
                d'8..
                e'32
                ~
                e'4
                e'4
                c'4
                d'8..
                e'32
                ~
                e'4
                f'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-AqHoQPgphf.png

    :attr:`total_duration`:
        The property :attr:`total_duration` can be used to check the total
        duration of the window size (that is, :attr:`duration_unit` times
        :attr:`units_per_window`).

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     duration_unit=abjad.Duration((1, 32)),
        ...     units_per_window=16,
        ... )
        >>> maker.total_duration
        1/2

    :attr:`omit_time_signature`:
        By default, a time signature is added to the output automatically:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'", "g'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12, 16],
        ...     attack_point_genes=list(range(20)),
        ...     units_per_window=20,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/4
                r8
                c'8
                d'8
                f'4.
                e'4..
                g'16
                c'8.
                d'16
                ~
                d'4
                ~
                d'16
                g'16
                f'8
                ~
                f'4
                g'4
                c'4
                d'8.
                e'16
                ~
                e'4
                a'4
                f'4
                c'4
                d'8.
                g'16
                ~
                g'4
                f'4
                g'4
                c'4
                d'4
                g'4
                f'4
                g'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-bbsRWshiDI.png

        Setting :attr:`omit_time_signature` to ``True`` will result in no time
        signature. Note that the output might need to be cleaned up using
        |abjad.Meter.rewrite_meter()|.

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'", "g'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12, 16],
        ...     attack_point_genes=list(range(20)),
        ...     units_per_window=20,
        ...     omit_time_signature=True,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                r8
                c'8
                d'8
                f'4.
                e'4..
                g'16
                c'8.
                d'4.
                g'16
                f'4.
                g'4
                c'4
                d'8.
                e'4
                ~
                e'16
                a'4
                f'4
                c'4
                d'8.
                g'4
                ~
                g'16
                f'4
                g'4
                c'4
                d'4
                g'4
                f'4
                g'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-IVSrzPV4OT.png

    :attr:`time_signatures`:
        Time signatures can also be enforced in the output. Set
        :attr:`time_signatures` to a single |abjad.TimeSignature| or a
        :obj:`list` of |abjad.TimeSignature|'s as needed.

        A single |abjad.TimeSignature| is applied to all measures:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     duration_unit=abjad.Duration((1, 32)),
        ...     time_signatures=abjad.TimeSignature((2, 2))
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/2
                r32
                c''16.
                ~
                c''32
                f'16.
                ~
                f'32
                e'16.
                f'8
                r2
                r32
                c'16.
                d'8
                ~
                d'16
                e'16
                e'8
                r2
                c'8.
                d'16
                e'8
                f'8
                r2
                r32
                c'16.
                d'8
                e'8
                f'8
                r2
                r32
                c'16.
                d'8
                e'8
                f'8
                r2
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-Xc9iKXJoEx.png

        A :obj:`list` of |abjad.TimeSignature|'s is applied cyclically.

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'", "g'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12, 16],
        ...     attack_point_genes=list(range(20)),
        ...     duration_unit=abjad.Duration((1, 16)),
        ...     units_per_window=20,
        ...     time_signatures=[abjad.TimeSignature((2, 4)),
        ...                      abjad.TimeSignature((3, 4)),
        ...                      ],
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                r8
                c'8
                d'8
                f'8
                ~
                \time 3/4
                f'4
                e'4..
                g'16
                \time 2/4
                c'8.
                d'16
                ~
                d'4
                ~
                \time 3/4
                d'16
                g'16
                f'4.
                g'4
                \time 2/4
                c'4
                d'8.
                e'16
                ~
                \time 3/4
                e'4
                a'4
                f'4
                \time 2/4
                c'4
                d'8.
                g'16
                ~
                \time 3/4
                g'4
                f'4
                g'4
                \time 2/4
                c'4
                d'4
                \time 3/4
                g'4
                f'4
                g'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-lJCYtjGY52.png

    :attr:`pitch_score_bias`:
        Pitches and attack points are scored separately and, by default,
        contribute equally to the total score of each individual. To change the
        bias of the pitch score, set :attr:`pitch_score_bias` to a value
        between ``0.0`` and ``1.0``.

        This is the default output:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                r16
                c'8
                d'16
                ~
                d'4
                f'4
                d'4
                r16
                c'8.
                d'8
                e'8
                ~
                e'8.
                f'16
                ~
                f'4
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-VEinjiTe3F.png

        With a high :attr:`pitch_score_bias`, pitch convergence will tend to be
        faster at the expense of attack points:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     pitch_score_weight=0.95,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                r8
                c'16
                d'16
                ~
                d'4
                ~
                d'8
                e'4
                d'8
                r8
                c'16
                d'16
                ~
                d'2
                ~
                d'16
                e'8.
                c'16
                d'8.
                ~
                d'8.
                e'16
                ~
                e'4
                f'4
                c'16
                d'8.
                ~
                d'8.
                e'16
                ~
                e'4
                f'4
                c'4
                ~
                c'16
                d'8
                e'16
                ~
                e'4
                f'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-iVD6l66hNl.png


        In contrast, a low :attr:`pitch_score_bias`, attack point convergence
        will tend to be  faster at the expense of pitches:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     pitch_score_weight=0.05,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'16
                c''4..
                f'4
                c''4
                e'4
                d'4
                d'4
                f'4
                e'4
                d'4
                d'4
                f'4
                e'4
                d'4
                d'4
                f'4
                e'4
                d'4
                d'4
                f'4
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-1Ex8sgcS2s.png

    :attr:`attack_points_mode`:
        When using this class in attack points mode, each note will last a
        single unit instead of being extended until the next attack point:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ...     attack_points_mode=True,
        ... )
        >>> notes = maker.output_n(5)
        >>> staff = abjad.Staff(notes)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                e'16
                r8.
                c'16
                r16
                r16
                g'16
                r4
                f'16
                r8.
                c'16
                r8.
                d'16
                r8.
                e'16
                c'16
                r4.
                c'16
                r8.
                d'16
                r16
                r16
                e'16
                r4
                f'16
                r8.
                c'16
                r8.
                d'16
                r8.
                e'16
                r8.
                f'16
                r8.
                c'16
                r8.
                d'16
                r8.
                e'16
                r8.
                f'16
                r8.
            }

        ..  figure:: ../_images/GeneticAlgorithmMusicMaker-5aInHHwMol.png

    :meth:`reset`:
        Use the :meth:`reset` method to reset the genetic algorithm at any
        point:

        >>> maker = auxjad.GeneticAlgorithmMusicMaker(
        ...     pitch_target=["c'", "d'", "e'", "f'"],
        ...     pitch_genes=["c'", "d'", "e'", "f'", "g'", "a'", "b'", "c''"],
        ...     attack_point_target=[0, 4, 8, 12],
        ...     attack_point_genes=list(range(16)),
        ... )
        >>> maker.generation_number
        None
        >>> maker.fittest_pitch_individual
        None
        >>> maker.fittest_attack_point_individual
        None
        >>> maker.fittest_individual_score
        None
        >>> for _ in range(10):
        ...     maker()
        >>> maker.generation_number
        9
        >>> maker.fittest_pitch_individual
        ["c'", "d'", "e'", "f'"]
        >>> maker.fittest_attack_point_individual
        [0, 4, 8, 12]
        >>> maker.fittest_individual_score
        0.5
        >>> maker.reset()
        >>> maker.generation_number
        None
        >>> maker.fittest_pitch_individual
        None
        >>> maker.fittest_attack_point_individual
        None
        >>> maker.fittest_individual_score
        None
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_pitch_target",
        "_pitch_genes",
        "_pitch_initial_individual",
        "_attack_point_target",
        "_attack_point_genes",
        "_attack_point_initial_individual",
        "_population_size",
        "_select_n_parents",
        "_mutation_chance",
        "_mutation_index",
        "_evaluation_index",
        "_duration_unit",
        "_units_per_window",
        "_omit_time_signature",
        "_time_signatures",
        "_attack_points_mode",
        "_pitch_score_bias",
        "_pitch_ga",
        "_attack_point_ga",
        "_fittest_measure",
        "_fittest_pitch_individual",
        "_fittest_attack_point_individual",
        "_pitch_population",
        "_attack_point_population",
        "_scores",
        "_target_music",
        "_total_duration",
    )

    ### INITIALISER ###

    def __init__(
        self,
        *,
        pitch_target: list,
        pitch_genes: list,
        attack_point_target: list,
        attack_point_genes: list,
        duration_unit: abjad.Duration = abjad.Duration((1, 16)),
        units_per_window: int = 16,
        pitch_initial_individual: Optional[list] = None,
        attack_point_initial_individual: Optional[list] = None,
        population_size: int = 100,
        select_n_parents: int = 10,
        keep_n_parents: int = 0,
        mutation_chance: float = 0.2,
        mutation_index: float = 0.1,
        evaluation_index: float = 0.2,
        omit_time_signature: bool = False,
        time_signatures: Optional[list] = None,
        attack_points_mode: bool = False,
        pitch_score_bias: float = 0.5,
    ) -> None:
        if len(pitch_target) != len(attack_point_target):
            raise ValueError("'pitch_target' and 'attack_point_target' must have the same length")
        self._pitch_ga = GeneticAlgorithm(
            target=pitch_target,
            genes=pitch_genes,
            initial_individual=pitch_initial_individual,
            population_size=population_size,
            select_n_parents=select_n_parents,
            keep_n_parents=keep_n_parents,
            mutation_chance=mutation_chance,
            mutation_index=mutation_index,
            evaluation_index=evaluation_index,
        )
        self._attack_point_ga = GeneticAlgorithm(
            target=attack_point_target,
            genes=attack_point_genes,
            initial_individual=attack_point_initial_individual,
            population_size=population_size,
            select_n_parents=select_n_parents,
            keep_n_parents=keep_n_parents,
            mutation_chance=mutation_chance,
            mutation_index=mutation_index,
            evaluation_index=evaluation_index,
        )
        if not isinstance(duration_unit, abjad.Duration):
            raise TypeError("'duration_unit' must be 'abjad.Duration'")
        if not isinstance(units_per_window, int):
            raise TypeError("'units_per_window' must be 'int'")
        if units_per_window < max(self._attack_point_ga.genes):
            raise TypeError(
                "'units_per_window' must be larger than the max value in 'attack_point_genes'"
            )
        self._duration_unit = duration_unit
        self._units_per_window = units_per_window
        self._total_duration = self._units_per_window * self._duration_unit
        self.omit_time_signature = omit_time_signature
        self.time_signatures = time_signatures
        self.attack_points_mode = attack_points_mode
        self.pitch_score_bias = pitch_score_bias
        self._target_individual_to_measure()

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r"""Returns interpreter representation of :attr:`target`'s of both
        instances of the genetic algorithm (pitches and attack points).
        """
        strings = (
            f"pitches: {repr(self._pitch_ga._target)}",
            f"attack_points: {repr(self._attack_point_ga._target)}",
        )
        return "\n".join(strings)

    def __len__(self) -> int:
        r"""Returns the number of genes in each individual."""
        return len(self._pitch_ga._target)

    def __call__(self) -> abjad.Selection:
        r"""Calls the genetic algorithm process for one iteration, returning an
        |abjad.Selection|. Generates a new generation of length
        :attr:`population_size` via reproduction and mutation processes and
        scores each individual using the evaluation function.
        """
        self._pitch_ga._generate_population()
        self._pitch_ga._score_population()
        self._attack_point_ga._generate_population()
        for attack_point_individual in self._attack_point_ga._population:
            attack_point_individual.sort()
        self._attack_point_ga._score_population()
        self._sort_population_by_evaluation()
        self._fittest_individual_to_measure()
        return self.fittest_measure

    def __next__(self) -> None:
        r"""Calls the genetic algorithm process for one iteration. Generates a
        new generation of length :attr:`population_size` via reproduction and
        mutation processes and scores each individual using the evaluation
        function.
        """
        try:
            return self.__call__()
        except RuntimeError:
            raise StopIteration

    def __iter__(self) -> None:
        r"""Returns an iterator, allowing instances to be used as iterators."""
        return self

    ### PUBLIC METHODS ###

    def reset(self) -> None:
        r"""Resets that genetic algorithm."""
        self._pitch_ga.reset()
        self._attack_point_ga.reset()
        self._pitch_population = None
        self._attack_point_population = None
        self._scores = None

    def output_n(self, n: int) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the genetic algorithm process and
        outputs a single |abjad.Selection|.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive 'int'")
        dummy_container = abjad.Container()
        for _ in range(n):
            dummy_container.append(self.__call__())
        mutate.remove_repeated_time_signatures(dummy_container[:])
        output = dummy_container[:]
        dummy_container[:] = []
        return output

    ### PRIVATE METHODS ###

    def _sort_population_by_evaluation(self) -> None:
        r"""Sorts the population (and their scores) according to the evaluation
        of its individuals.
        """
        self._scores = []
        for pitch_score, attack_score in zip(self._pitch_ga._scores, self._attack_point_ga._scores):
            combined_score = pitch_score * self._pitch_score_bias
            combined_score += attack_score * (1.0 - self._pitch_score_bias)
            combined_score /= 2
            self._scores.append(combined_score)
        zipped_lists = list(
            zip(
                self._scores,
                self._pitch_ga._population,
                self._attack_point_ga._population,
            )
        )
        zipped_lists.sort(
            key=lambda tuple_triple: tuple_triple[0],
            reverse=True,
        )
        self._pitch_population = [pitch_gene for _, pitch_gene, _ in zipped_lists]
        self._attack_point_population = [
            attack_point_gene for _, _, attack_point_gene in zipped_lists
        ]
        self._scores = [score for score, _, _ in zipped_lists]
        self._pitch_ga._population = self._pitch_population[:]
        self._attack_point_ga._population = self._attack_point_population[:]

    def _fittest_individual_to_measure(self) -> None:
        r"""Converts the fittest pitch and attack point individuals to a
        measure of music.
        """
        self._fittest_measure = self._make_measure(
            self.fittest_attack_point_individual[:],
            self.fittest_pitch_individual[:],
        )

    def _target_individual_to_measure(self) -> None:
        r"""Converts the target pitch and attack point individuals to a
        measure of music.
        """
        self._target_music = self._make_measure(
            self._attack_point_ga.target[:],
            self._pitch_ga.target[:],
        )

    @staticmethod
    def _sort_by_attack_point(
        attack_points: list,
        pitches: list,
    ) -> tuple:
        r"""Sorts pitches and attack points."""
        zipped_lists = []
        # zipping while removing simultaneous attacks
        for attack_point, pitch in zip(attack_points, pitches):
            if attack_point not in (p for p, _ in zipped_lists):
                zipped_lists.append((attack_point, pitch))
        zipped_lists.sort(
            key=lambda tuple_pair: tuple_pair[0],
        )
        attack_points = [attack_point for attack_point, _ in zipped_lists]
        pitches = [pitch for _, pitch in zipped_lists]
        return attack_points, pitches

    def _convert_attack_points_to_durations(
        self,
        attack_points: list,
        pitches: list,
    ) -> tuple:
        r"""Converts attack points to effective durations. Adds an initial rest
        if first attack point is not at the 0-th position.
        """
        if attack_points[0] != 0:
            attack_points.insert(0, 0)
            pitches.insert(0, None)
        durations = []
        attack_points.append(self._units_per_window)
        for i in range(len(attack_points) - 1):
            difference = attack_points[i + 1] - attack_points[i]
            duration = difference * self.duration_unit
            durations.append(duration)
        return durations, pitches

    def _make_measure(
        self,
        attack_points,
        pitches,
    ) -> None:
        r"""Converts a list of pitch and attack point individuals into a
        measure of music.
        """
        dummy_container = abjad.Container()
        sorted_attack_points, sorted_pitches = self._sort_by_attack_point(
            attack_points[:],
            pitches[:],
        )
        if not self._attack_points_mode:
            durations, pitches = self._convert_attack_points_to_durations(
                sorted_attack_points,
                sorted_pitches,
            )
            notes = abjad.LeafMaker()(
                pitches,
                durations,
            )
            dummy_container = abjad.Container(notes)
        else:
            notes = abjad.Selection()
            for i in range(self._units_per_window):
                if i in sorted_attack_points:
                    pitch_index = sorted_attack_points.index(i)
                    notes += abjad.LeafMaker()(
                        [sorted_pitches[pitch_index]],
                        [self._duration_unit],
                    )
                else:
                    rest = abjad.Rest(self._duration_unit)
                    notes += abjad.select(rest).leaves()
            dummy_container = abjad.Container(notes)
        # adding time signature
        if not self._omit_time_signature:
            if self._time_signatures is None:
                time_signature = abjad.TimeSignature((self._total_duration))
                time_signature.simplify_ratio()
                abjad.attach(
                    time_signature,
                    dummy_container[0],
                )
                mutate.auto_rewrite_meter(dummy_container)
            else:
                mutate.enforce_time_signature(
                    dummy_container,
                    self._time_signatures,
                )
        return dummy_container[:]

    ### PUBLIC PROPERTIES ###

    @property
    def duration_unit(self) -> abjad.Duration:
        r"""Unit for the duration grid. Default is abjad.Duration((1, 16))."""
        return self._duration_unit

    @duration_unit.setter
    def duration_unit(
        self,
        duration_unit: abjad.Duration,
    ) -> None:
        if not isinstance(duration_unit, abjad.Duration):
            raise TypeError("'duration_unit' must be 'abjad.Duration'")
        self._duration_unit = duration_unit
        self._total_duration = self._units_per_window * self._duration_unit

    @property
    def units_per_window(self) -> int:
        r"""Number of units per window. Default is 16."""
        return self._units_per_window

    @units_per_window.setter
    def units_per_window(
        self,
        units_per_window: int,
    ) -> None:
        if not isinstance(units_per_window, int):
            raise TypeError("'units_per_window' must be 'int'")
        if units_per_window < max(self._attack_point_ga.genes):
            raise TypeError(
                "'units_per_window' must be larger than the max value in 'attack_point_genes'"
            )
        self._units_per_window = units_per_window
        self._total_duration = self._units_per_window * self._duration_unit

    @property
    def omit_time_signature(self) -> bool:
        r"""When ``True``, a time signature won't be added to the first leaf of
        the output.
        """
        return self._omit_time_signature

    @omit_time_signature.setter
    def omit_time_signature(
        self,
        omit_time_signature: bool,
    ) -> None:
        if not isinstance(omit_time_signature, bool):
            raise TypeError("'omit_time_signature' must be 'bool'")
        self._omit_time_signature = omit_time_signature

    @property
    def time_signatures(self) -> list:
        r"""List of time signatures to be enforced on output. It is important
        to note that :attr:`omit_time_signature` must be ``True`` for it to
        take effect.
        """
        return self._time_signatures

    @time_signatures.setter
    def time_signatures(
        self,
        time_signatures: Optional[list],
    ) -> None:
        if time_signatures is not None:
            if isinstance(time_signatures, abjad.TimeSignature):
                time_signatures = [time_signatures]
            elif not isinstance(time_signatures, list):
                raise TypeError(
                    "'time_signatures' must be a 'list' of 'abjad.TimeSignature' or 'None'"
                )
            if not all(isinstance(ts, abjad.TimeSignature) for ts in time_signatures):
                raise TypeError(
                    "'time_signatures' must be a 'list' of 'abjad.TimeSignature' or 'None'"
                )
        self._time_signatures = time_signatures

    @property
    def attack_points_mode(self) -> bool:
        r"""When ``True``, each note will last only for the duration of the
        unit, instead of extending it to the next attack point.
        """
        return self._attack_points_mode

    @attack_points_mode.setter
    def attack_points_mode(
        self,
        attack_points_mode: bool,
    ) -> None:
        if not isinstance(attack_points_mode, bool):
            raise TypeError("'attack_points_mode' must be 'bool'")
        self._attack_points_mode = attack_points_mode

    @property
    def pitch_target(self) -> list:
        r"""Target pitch individual used for evaluation."""
        return self._pitch_ga.target

    @pitch_target.setter
    def pitch_target(
        self,
        pitch_target: list,
    ) -> None:
        if len(pitch_target) != len(self._attack_point_ga.target):
            raise ValueError("'pitch_target' must be the same length as 'attack_point_target'")
        self._pitch_ga.target = pitch_target
        self._target_individual_to_measure()

    @property
    def pitch_genes(self) -> list:
        r"""List of possible genes that make up all pitch individuals."""
        return self._pitch_ga.genes

    @pitch_genes.setter
    def pitch_genes(
        self,
        pitch_genes: list,
    ) -> None:
        self._pitch_ga.genes = pitch_genes

    @property
    def pitch_initial_individual(self) -> Union[list, None]:
        r"""Optional initial pitch individual."""
        return self._pitch_ga.initial_individual

    @pitch_initial_individual.setter
    def pitch_initial_individual(
        self,
        pitch_initial_individual: Optional[list],
    ) -> None:
        self._pitch_ga.initial_individual = pitch_initial_individual

    @property
    def attack_point_target(self) -> list:
        r"""Target attack point individual used for evaluation."""
        return self._attack_point_ga.target

    @attack_point_target.setter
    def attack_point_target(
        self,
        attack_point_target: list,
    ) -> None:
        if len(attack_point_target) != len(self._pitch_ga.target):
            raise ValueError("'attack_point_target' must be the same length as 'pitch_target'")
        self._attack_point_ga.target = attack_point_target
        self._target_individual_to_measure()

    @property
    def attack_point_genes(self) -> list:
        r"""List of possible genes that make up all attack point
        individuals.
        """
        return self._attack_point_ga.genes

    @attack_point_genes.setter
    def attack_point_genes(
        self,
        attack_point_genes: list,
    ) -> None:
        self._attack_point_ga.genes = attack_point_genes

    @property
    def attack_point_initial_individual(self) -> Union[list, None]:
        r"""Optional initial attack point individual."""
        return self._attack_point_ga.initial_individual

    @attack_point_initial_individual.setter
    def attack_point_initial_individual(
        self,
        attack_point_initial_individual: Optional[list],
    ) -> None:
        self._attack_point_ga.initial_individual = attack_point_initial_individual

    @property
    def population_size(self) -> int:
        r"""Number of individuals in any given generation."""
        return self._pitch_ga.population_size

    @population_size.setter
    def population_size(
        self,
        population_size: int,
    ) -> None:
        self._pitch_ga.population_size = population_size
        self._attack_point_ga.population_size = population_size

    @property
    def select_n_parents(self) -> int:
        r"""Number of the best-fit individuals that are selected to be parents
        of the next generation. They also survive into the next generation.
        """
        return self._pitch_ga.select_n_parents

    @select_n_parents.setter
    def select_n_parents(
        self,
        select_n_parents: int,
    ) -> None:
        self._pitch_ga.select_n_parents = select_n_parents
        self._attack_point_ga.select_n_parents = select_n_parents

    @property
    def keep_n_parents(self) -> int:
        r"""Number of the best-fit individuals that survive into the next
        generation. Default is ``0``.
        """
        return self._pitch_ga.keep_n_parents

    @keep_n_parents.setter
    def keep_n_parents(
        self,
        keep_n_parents: int,
    ) -> None:
        self._pitch_ga.keep_n_parents = keep_n_parents
        self._attack_point_ga.keep_n_parents = keep_n_parents

    @property
    def mutation_chance(self) -> float:
        r"""Percentage of the total population who will experience mutation."""
        return self._pitch_ga.mutation_chance

    @mutation_chance.setter
    def mutation_chance(
        self,
        mutation_chance: float,
    ) -> None:
        self._pitch_ga.mutation_chance = mutation_chance
        self._attack_point_ga.mutation_chance = mutation_chance

    @property
    def mutation_index(self) -> float:
        r"""Given an individual selected to undergo mutation, this index gives
        the percentage of genes of that individual which will be mutated.
        """
        return self._pitch_ga.mutation_index

    @mutation_index.setter
    def mutation_index(
        self,
        mutation_index: float,
    ) -> None:
        self._pitch_ga.mutation_index = mutation_index
        self._attack_point_ga.mutation_index = mutation_index

    @property
    def evaluation_index(self) -> float:
        r"""The index used in the evaluation function. This index will be
        raised by the difference between indices of the target value and the
        current value. Consider the following example, where the available
        genes are ``['A', 'B', 'C', 'D', 'E', 'F']`` and the target is
        ``['B', 'A', 'A', 'C']``. Suppose an individual has the genes
        ``['D', 'D', 'A', 'B']``.

        To evaluate this individual, first the algorithm finds the indices of
        both the target's genes (in this case, ``[1, 0, 0, 2]``) and also of
        the individual to be evaluated (in this case, ``[3, 3, 0, 1]``). It
        then scores each element of this individual against the target using:

        .. code-block::

            difference = abs(target_gene_index - individual_gene_index)
            element_score = evaluation_index ** difference

        Thus, when the difference is ``0``, the score of this element is
        ``1.0``. The higher the difference, the smaller the value; this decay
        can be controlled using this very property, whose default value is
        ``0.2``. Thus when the difference is ``1`` or ``-1``, the score is
        ``0.2 ** 1 = 0.2``, when the difference is ``2`` or ``-2``, the score
        is ``0.2 ** 2 = 0.04``, and so on. The total score of an individual
        will be given by the normalised sum of the evaluation of each of its
        genes.
        """
        return self._pitch_ga.evaluation_index

    @evaluation_index.setter
    def evaluation_index(
        self,
        evaluation_index: float,
    ) -> None:
        self._pitch_ga.evaluation_index = evaluation_index
        self._attack_point_ga.evaluation_index = evaluation_index

    @property
    def pitch_score_bias(self) -> float:
        r"""By default, the score of each measure gives equal weight to pitches
        as it gives to attack points. Changing this to a different value will
        make the pitch score contribute more or less to the total score of a
        measure.
        """
        return self._pitch_score_bias

    @pitch_score_bias.setter
    def pitch_score_bias(
        self,
        pitch_score_bias: float,
    ) -> None:
        if not isinstance(pitch_score_bias, float):
            raise TypeError("'pitch_score_bias' must be 'float'")
        if pitch_score_bias < 0.0 or pitch_score_bias > 1.0:
            raise ValueError("'pitch_score_bias' must be between 0.0 and 1.0")
        self._pitch_score_bias = pitch_score_bias

    @property
    def fittest_measure(self) -> Union[abjad.Selection, None]:
        r"""Read-only property, returns the fittest individual of the current
        population as an |abjad.Selection|.
        """
        return abjad.mutate.copy(self._fittest_measure)

    @property
    def target_music(self) -> abjad.Selection:
        r"""Read-only property, returns the target as an |abjad.Selection|."""
        return abjad.mutate.copy(self._target_music)

    @property
    def total_duration(self) -> list:
        r"""Read-only property, returns the total duration of the window."""
        return self._total_duration

    @property
    def generation_number(self) -> list:
        r"""Read-only property, returns the number of the current generation
        (initial generation is ``0``).
        """
        return self._pitch_ga._generation_number

    @property
    def pitch_population(self) -> Union[list, None]:
        r"""Read-only property, returns a list with all the population of the
        current generation.
        """
        return self._pitch_population

    @property
    def attack_point_population(self) -> Union[list, None]:
        r"""Read-only property, returns a list with all the population of the
        current generation.
        """
        return self._attack_point_population

    @property
    def scores(self) -> list:
        r"""Read-only property, returns the list of individual scores of the
        current population. Scores are normalised.
        """
        return self._scores

    @property
    def fittest_pitch_individual(self) -> Union[list, None]:
        r"""Read-only property, returns the fittest individual of the current
        population.
        """
        try:
            return self._pitch_population[0]
        except AttributeError:
            return None
        except TypeError:
            return None

    @property
    def fittest_attack_point_individual(self) -> Union[list, None]:
        r"""Read-only property, returns the fittest individual of the current
        population.
        """
        try:
            return self._attack_point_population[0]
        except AttributeError:
            return None
        except TypeError:
            return None

    @property
    def fittest_individual_score(self) -> Union[list, float]:
        r"""Read-only property, returns the score of the fittest individual of
        the current population.
        """
        try:
            return self._scores[0]
        except AttributeError:
            return None
        except TypeError:
            return None



import abjad

from .remove_repeated_dynamics import (
    remove_repeated_dynamics as remove_repeated_dynamics_,
)


def reposition_dynamics(
    selection: abjad.Selection,
    *,
    allow_hairpins_under_rests: bool = False,
    check_hairpin_trends: bool = True,
    remove_repeated_dynamics: bool = True,
    allow_hairpin_to_rest_with_dynamic: bool = True,
) -> None:
    r"""Mutates an input |abjad.Selection| in place and has no return value;
    this function shifts all dynamics from rests to the next pitched leaves. It
    will also adjust hairpins if necessary.

    Basic usage:
        This function will shift dynamics under rests to the next pitched leaf.

        >>> staff = abjad.Staff(r"c'1\p d'2 r2\f r1 e'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'2
                r2
                \f
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_dynamics-sqwvevg7o9.png

        >>> staff = abjad.Staff(r"c'1\p d'2 r2\f r1 e'1")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'2
                r2
                r1
                e'1
                \f
            }

        ..  figure:: ../_images/reposition_dynamics-v18uzh1zjs.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.mutate.reposition_dynamics(staff[:])

    Removing dynamics:
        If the next pitched leaf already contain a dynamic, this function will
        simply remove the dynamic under the rest.

        >>> staff = abjad.Staff(r"c'1\p d'2 r2\f r1\mf e'1\pp")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'2
                r2
                \f
                r1
                \mf
                e'1
                \pp
            }

        ..  figure:: ../_images/reposition_dynamics-aom2qywcn9m.png

        >>> staff = abjad.Staff(r"c'1\p d'2 r2\f r1\mf e'1\pp")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'2
                r2
                r1
                e'1
                \pp
            }

        ..  figure:: ../_images/reposition_dynamics-2ua73x102fp.png

    ``remove_repeated_dynamics``:
        By default indentical repeated dynamics are omitted.

        >>> staff = abjad.Staff(r"c'1\p d'1 r1\f e'1\p")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'1
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_dynamics-i4x8b1z1ak.png

        Set the optional keyword argument ``remove_repeated_dynamics`` to
        ``False`` to disable this behaviour.

        >>> staff = abjad.Staff(r"c'1\p d'1 r1\f e'1\p")
        >>> auxjad.mutate.reposition_dynamics(
        ...     staff[:],
        ...     remove_repeated_dynamics=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'1
                r1
                e'1
                \p
            }

        ..  figure:: ../_images/reposition_dynamics-se8a5aeqer.png

    ``allow_hairpins_under_rests``:
        This function will shorten hairpins until rests by default.

        >>> staff = abjad.Staff(r"c'1\p\< d'2 r2 r1\f e'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'2
                r2
                r1
                \f
                e'1
            }

        ..  figure:: ../_images/reposition_dynamics-r28po3j3hd.png

        >>> staff = abjad.Staff(r"c'1\p\< d'2 r2 r1\f e'1")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'2
                r2
                \!
                r1
                e'1
                \f
            }

        ..  figure:: ../_images/reposition_dynamics-n60vvnqrcnr.png

        Set the optional keyword argument ``allow_hairpins_under_rests`` to
        ``True`` to allow hairpins to extend cross rests.

        >>> staff = abjad.Staff(r"c'1\p\< d'2 r2 r1\f e'1")
        >>> auxjad.mutate.reposition_dynamics(
        ...     staff[:],
        ...     allow_hairpins_under_rests=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'2
                r2
                r1
                e'1
                \f
            }

        ..  figure:: ../_images/reposition_dynamics-ugit7ijz89h.png

    ``allow_hairpin_to_rest_with_dynamic``:
        Notice that if a hairpin leads to a rest with dynamic, that one is not
        removed.

        >>> staff = abjad.Staff(r"c'1\p\< d'2 r2\f r1 e'1")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'2
                r2
                \f
                r1
                e'1
            }

        ..  figure:: ../_images/reposition_dynamics-t9z4y5zzj6.png

        Set the argument ``allow_hairpin_to_rest_with_dynamic`` to ``False`` to
        disable this behaviour.

        >>> staff = abjad.Staff(r"c'1\p\< d'2 r2\f r1 e'1")
        >>> auxjad.mutate.reposition_dynamics(
        ...     staff[:],
        ...     allow_hairpin_to_rest_with_dynamic=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \<
                d'2
                r2
                \!
                r1
                e'1
                \f
            }

        ..  figure:: ../_images/reposition_dynamics-wpwweov55qf.png

    ``check_hairpin_trends``:
        This function will remove any hairpins connecting dynamics that grow in
        the opposite direction to the hairpin's trend, such as a diminuendo
        hairpin from piano to forte.

        >>> staff = abjad.Staff(r"c'1\p\> d'1\f\> e'1\p")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                d'1
                \f
                \>
                e'1
                \p
            }

        ..  figure:: ../_images/reposition_dynamics-f0b4ppb71ii.png

        This behaviour can be disabled by setting the argument
        ``check_hairpin_trends`` to ``False``.

        >>> staff = abjad.Staff(r"c'1\p\> d'1\f\> e'1\p")
        >>> auxjad.mutate.reposition_dynamics(
        ...     staff[:],
        ...     check_hairpin_trends=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \>
                d'1
                \f
                \>
                e'1
                \p
            }

        ..  figure:: ../_images/reposition_dynamics-1e2ugszm95fi.png

    ..  note::

        The behaviour described above is only applicable when a hairpin ends
        on a dynamic. Using the hairpin terminator ``\!`` before a dynamic
        change will not cause a hairpin to be removed as it is not considered
        to be connecting dynamics of the opposite trend.

        >>> staff = abjad.Staff(r"c'1\p\> d'1\! e'1\f\> f'1\p")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                \>
                d'1
                \!
                e'1
                \f
                \>
                f'1
                \p
            }

        ..  figure:: ../_images/reposition_dynamics-77g0uwthbgd.png

    Types of hairpins:
        This function can handle multiple types of hairpins as well as niente
        dynamics.

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 r1\mf r1\ff f'1 r1 g'1")
        >>> abjad.attach(abjad.Dynamic('niente', hide=True), staff[0])
        >>> abjad.attach(abjad.Dynamic('niente', hide=True), staff[7])
        >>> abjad.attach(abjad.StartHairpin('o<'), staff[0])
        >>> abjad.attach(abjad.StartHairpin('>o'), staff[4])
        >>> abjad.attach(abjad.StopHairpin(), staff[7])
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                - \tweak circled-tip ##t
                \<
                d'1
                e'1
                r1
                \mf
                r1
                f'1
                \ff
                - \tweak circled-tip ##t
                \>
                r1
                \!
                g'1
            }

        ..  figure:: ../_images/reposition_dynamics-m8it9awv1ce.png

        >>> staff = abjad.Staff(
        ...     r"c'1\p d'1\f\> e'1\ff\< r1\fff f'1\p\> g'1\ppp"
        ... )
        >>> abjad.attach(abjad.StartHairpin('--'), staff[0])
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                - \tweak stencil #constante-hairpin
                \<
                d'1
                \f
                e'1
                \ff
                \<
                r1
                \fff
                f'1
                \p
                \>
                g'1
                \ppp
            }

        ..  figure:: ../_images/reposition_dynamics-fosad5ltzj.png

    Multi-measure rests:
        Multi-measure rests are also supported.

        >>> staff = abjad.Staff(r"c'1\p R1\f d'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                R1
                \f
                d'1
            }

        ..  figure:: ../_images/reposition_dynamics-uj6jasfs2uh.png

        >>> staff = abjad.Staff(r"c'1\p R1\f d'1")
        >>> auxjad.mutate.reposition_dynamics(staff[:])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                \p
                R1
                d'1
                \f
            }

        ..  figure:: ../_images/reposition_dynamics-axpcbm9hocd.png

    ..  warning::

        The input selection must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(selection, abjad.Selection):
        raise TypeError("argument must be 'abjad.Selection'")
    if not abjad.select(selection).leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    if not isinstance(allow_hairpins_under_rests, bool):
        raise TypeError("'allow_hairpins_under_rests' must be 'bool'")
    if not isinstance(check_hairpin_trends, bool):
        raise TypeError("'check_hairpin_trends' must be 'bool'")
    if not isinstance(remove_repeated_dynamics, bool):
        raise TypeError("'remove_repeated_dynamics' must be 'bool'")
    if not isinstance(allow_hairpin_to_rest_with_dynamic, bool):
        raise TypeError("'allow_hairpin_to_rest_with_dynamic' must be 'bool'")

    leaves = selection.leaves()

    # shifting dynamics and hairpins from rests to notes
    shifted_dynamic = None
    shifted_hairpin = None
    active_hairpin = None
    for leaf in leaves:
        if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)):
            if abjad.get.indicator(leaf, abjad.Dynamic) is not None:
                previous_leaf = abjad.select(leaf).with_previous_leaf()[0]
                if (
                    allow_hairpin_to_rest_with_dynamic
                    and active_hairpin is not None
                    and not isinstance(
                        previous_leaf,
                        (abjad.Rest, abjad.MultimeasureRest),
                    )
                ):
                    active_hairpin = None
                else:
                    shifted_dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
                    abjad.detach(abjad.Dynamic, leaf)
            if abjad.get.indicator(leaf, abjad.StartHairpin) is not None:
                shifted_hairpin = abjad.get.indicator(leaf, abjad.StartHairpin)
                abjad.detach(abjad.StartHairpin, leaf)
        else:
            if abjad.get.indicator(leaf, abjad.Dynamic) is None:
                if shifted_dynamic is not None:
                    abjad.attach(shifted_dynamic, leaf)
                    if abjad.get.indicator(leaf, abjad.StopHairpin) is not None:
                        abjad.detach(abjad.StopHairpin, leaf)
                if shifted_hairpin is not None:
                    abjad.attach(shifted_hairpin, leaf)
            else:
                active_hairpin = None
            if abjad.get.indicator(leaf, abjad.StopHairpin) is not None:
                active_hairpin = None
            shifted_dynamic = None
            shifted_hairpin = None
            if active_hairpin is None:
                active_hairpin = abjad.get.indicator(leaf, abjad.StartHairpin)

    # stopping hairpins under rests if not allowed
    if not allow_hairpins_under_rests:
        effective_hairpin = None
        for leaf in leaves:
            start_hairpin = abjad.get.indicator(leaf, abjad.StartHairpin)
            if start_hairpin is not None:
                effective_hairpin = start_hairpin
                continue
            if isinstance(leaf, (abjad.Rest, abjad.MultimeasureRest)):
                if effective_hairpin is not None:
                    if abjad.get.indicator(leaf, abjad.StopHairpin) is None:
                        abjad.attach(abjad.StopHairpin(), leaf)
                    effective_hairpin = None
            else:
                dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
                stop_hairpin = abjad.get.indicator(leaf, abjad.StopHairpin)
                if dynamic is not None or stop_hairpin is not None:
                    effective_hairpin = None

    # cleaning up hairpins
    effective_dynamic = None
    for index, leaf in enumerate(leaves[:-1]):
        if abjad.get.indicator(leaf, abjad.Dynamic) is not None:
            effective_dynamic = abjad.get.indicator(leaf, abjad.Dynamic)
        start_hairpin = abjad.get.indicator(leaf, abjad.StartHairpin)
        if start_hairpin is not None and check_hairpin_trends:
            for next_leaf in leaves[index + 1 :]:
                next_dynamic = abjad.get.indicator(next_leaf, abjad.Dynamic)
                if next_dynamic is not None and effective_dynamic is not None:
                    if "<" in start_hairpin.shape:
                        if next_dynamic.ordinal <= effective_dynamic.ordinal:
                            abjad.detach(abjad.StartHairpin, leaf)
                    elif ">" in start_hairpin.shape:
                        if next_dynamic.ordinal >= effective_dynamic.ordinal:
                            abjad.detach(abjad.StartHairpin, leaf)
                    break
                elif abjad.get.indicator(next_leaf, abjad.StopHairpin) is not None:
                    break
    if abjad.get.indicator(leaves[-1], abjad.StartHairpin) is not None:
        abjad.detach(abjad.StartHairpin, leaves[-1])

    # removing unecessary StopHairpin's
    for leaf in leaves:
        if (
            abjad.get.indicator(leaf, abjad.StopHairpin) is not None
            and abjad.get.indicator(leaf, abjad.Dynamic) is not None
        ):
            abjad.detach(abjad.StopHairpin, leaf)
    target_leaf = None
    for leaf in leaves[::-1]:
        if abjad.get.indicator(leaf, abjad.StopHairpin) is not None:
            if target_leaf is not None:
                abjad.detach(abjad.StopHairpin, target_leaf)
            target_leaf = leaf
        elif (
            abjad.get.indicator(leaf, abjad.StartHairpin) is not None
            or abjad.get.indicator(leaf, abjad.Dynamic)
        ) is not None:
            target_leaf = None

    # removing repeated dynamics if required
    if remove_repeated_dynamics:
        remove_repeated_dynamics_(selection)



from typing import Optional, Union

import abjad

from ._HarmonicParent import _HarmonicParent


class ArtificialHarmonic(abjad.Chord, _HarmonicParent):
    r"""Creates a chord with a tweaked top note head for notating artificial
    harmonics. This is a child class of |abjad.Chord|.

    Basic usage:
        Usage is similar to |abjad.Chord|:

        >>> harm = auxjad.ArtificialHarmonic("<g c'>4")
        >>> harm.style
        "#'harmonic"
        >>> abjad.show(harm)

        ..  docs::

            <
                g
                \tweak style #'harmonic
                c'
            >4

        ..  figure:: ../_images/ArtificialHarmonic-16am9cj6p9u.png

        And similarly to |abjad.Chord|, pitch and duration can be input in
        many different ways:

        >>> harm1 = auxjad.ArtificialHarmonic(r"<g c'>4")
        >>> harm2 = auxjad.ArtificialHarmonic(["g", "c'"], 1 / 4)
        >>> harm3 = auxjad.ArtificialHarmonic([-5, 0], 0.25)
        >>> harm4 = auxjad.ArtificialHarmonic([-5, 0], abjad.Duration(1, 4))
        >>> staff = abjad.Staff([harm1, harm2, harm3, harm4])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <
                    g
                    \tweak style #'harmonic
                    c'
                >4
                <
                    g
                    \tweak style #'harmonic
                    c'
                >4
                <
                    g
                    \tweak style #'harmonic
                    c'
                >4
                <
                    g
                    \tweak style #'harmonic
                    c'
                >4
            }

        ..  figure:: ../_images/ArtificialHarmonic-je277bmakgs.png

    ..  error::
        It is important to note that this class can only be initialised with
        exactly two pitches. Any other number of pitches will raise a
        :exc:`ValueError`:

        >>> auxjad.ArtificialHarmonic(r"<g c' d'>4")
        ValueError: 'ArtificialHarmonic' requires exactly two
        'written_pitches' for initialisation

    :attr:`style`:
        When instantiating this class, use the keyword argument :attr:`style`
        to set a different type of note head for the top note, such as
        ``"#'harmonic-mixed"``:

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4",
        ...                                  style="#'harmonic-mixed",
        ...                                  )
        >>> harm.style
        "#'harmonic-mixed"
        >>> abjad.show(harm)

        ..  docs::

            <
                g
                \tweak style #'harmonic-mixed
                c'
            >4

        ..  figure:: ../_images/ArtificialHarmonic-ohqb65228iq.png

    :attr:`is_parenthesized`:
        To notate natural harmonics with a parenthesised pitch for the open
        string at the bottom of the interval, set the keyword
        :attr:`is_parenthesized` to ``True``.

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4",
        ...                                  is_parenthesized=True,
        ...                                  )
        >>> harm.is_parenthesized
        True
        >>> abjad.show(harm)

        ..  docs::

            <
                \parenthesize
                \tweak Parentheses.font-size -3
                g
                \tweak style #'harmonic
                c'
            >4

        ..  figure:: ../_images/ArtificialHarmonic-2q3jkx33yvl.png

    :attr:`~abjad.core.Chord.Chord.multiplier`:
        Similarly to |abjad.Chord|, this class can take multipliers:

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4",
        ...                                  multiplier=(2, 3),
        ...                                  )
        >>> harm.multiplier
        abjad.Multiplier(2, 3)
        >>> abjad.show(harm)

        ..  docs::

            <
                g
                \tweak style #'harmonic
                c'
            >4 * 2/3

        ..  figure:: ../_images/ArtificialHarmonic-ouhdk3ugkcs.png

    Properties:
        All properties of |abjad.Chord| are also available to be read. This
        class also includes two new properties named :attr:`style` and
        :attr:`is_parenthesized`:

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4")
        >>> harm.written_pitches
        "g c'"
        >>> harm.written_duration
        1/4
        >>> harm.style
        "#'harmonic"
        >>> harm.is_parenthesized
        False

        All these properties can be set to different values after
        initialisation:

        >>> harm.written_pitches = [-5, 2]
        >>> harm.written_duration = abjad.Duration(1, 8)
        >>> harm.style = "#'harmonic-mixed"
        >>> harm.is_parenthesized = True
        >>> harm.written_pitches
        "g d'"
        >>> harm.written_duration
        1/8
        >>> harm.style
        "#'harmonic-mixed"
        >>> harm.is_parenthesized
        True

    :meth:`sounding_pitch` and :meth:`sounding_note`:
        The methods :meth:`sounding_pitch` and :meth:`sounding_note` return the
        sounding pitch and sounding note, respectively. Their types are
        |abjad.Pitch| and |abjad.Note|, respectively.

        >>> harmonics = [ArtificialHarmonic(r"<g b>4"),
        ...              ArtificialHarmonic(r"<g c'>4"),
        ...              ArtificialHarmonic(r"<g d'>4"),
        ...              ArtificialHarmonic(r"<g e'>4"),
        ...              ArtificialHarmonic(r"<g g'>4"),
        ...              ]
        >>> for harmonic in harmonics:
        ...     print(harmonic.sounding_pitch())
        b''
        g''
        d''
        b''
        g'
        >>> for harmonic in harmonics:
        ...     print(harmonic.sounding_note())
        b''4
        g''4
        d''4
        b''4
        g'4

    :meth:`sounding_note` and indicators:
        The note created by :meth:`sounding_note` inherits all indicators of
        the artificial harmonic.

        >>> harm = auxjad.ArtificialHarmonic(r"<g c'>4-.\pp")
        >>> abjad.show(harm.sounding_note())

        ..  docs::

            g''4
            \pp
            - \staccato

        ..  figure:: ../_images/ArtificialHarmonic-dfabdv155mu.png

    ..  error::

        Both :meth:`sounding_pitch` and :meth:`sounding_note` methods raise a
        :exc:`ValueError` exception when it cannot calculate the sounding pitch
        for the given interval.

        >>> harm = auxjad.ArtificialHarmonic(r"<g ef'>4")
        >>> harm.sounding_pitch()
        ValueError: cannot calculate sounding pitch for given interval
        >>> harm.sounding_note()
        ValueError: cannot calculate sounding pitch for given interval

    :attr:`markup`:
        To add a markup expression to the artificial harmonic, use the
        :attr:`markup` optional keyword argument, which takes strings. By
        default, the markup position is above the harmonic note, but this can
        be overridden using the keyword :attr:`direction`, which can take
        strings as well as ``abjad.Up`` and ``abjad.Down``:

        >>> harm1 = auxjad.ArtificialHarmonic(r"<a d'>1")
        >>> harm2 = auxjad.ArtificialHarmonic(r"<a d'>1",
        ...                                   markup='I.',
        ...                                   )
        >>> harm3 = auxjad.ArtificialHarmonic(r"<a d'>1",
        ...                                   markup='I.',
        ...                                   direction=abjad.Down)
        >>> staff = abjad.Staff([harm1, harm2, harm3])
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <
                    a
                    \tweak style #'harmonic
                    d'
                >1
                \once \override TextScript.parent-alignment-X = 0
                \once \override TextScript.self-alignment-X = 0
                <
                    a
                    \tweak style #'harmonic
                    d'
                >1
                ^ \markup { I. }
                \once \override TextScript.parent-alignment-X = 0
                \once \override TextScript.self-alignment-X = 0
                <
                    a
                    \tweak style #'harmonic
                    d'
                >1
                _ \markup { I. }
            }

        ..  figure:: ../_images/ArtificialHarmonic-teysjphrrpn.png

        Setting :attr:`markup` to ``None`` will remove the markup from the
        note.

        >>> harm = auxjad.ArtificialHarmonic(r"<a d'>1",
        ...                                  markup='I.',
        ...                                  )
        >>> harm.markup = None
        >>> abjad.show(harm)

        ..  docs::

            <
                a
                \tweak style #'harmonic
                d'
            >1

        ..  figure:: ../_images/ArtificialHarmonic-nov336z64r.png

    :attr:`centre_markup`:
        When a markup expression is added to the harmonic note by using the
        :attr:`markup` optional keyword argument, it will be automatically
        centred above the note (as the main purpose of this markup is to show
        string numbers). To disable this behaviour, set :attr:`centre_markup`
        to ``False``. Compare:

        >>> harm1 = auxjad.ArtificialHarmonic(r"<a d'>1",
        ...                                   markup='III.',
        ...                                   )
        >>> abjad.show(harm1)

        ..  docs::

            \once \override TextScript.parent-alignment-X = 0
            \once \override TextScript.self-alignment-X = 0
            <
                bf'
                \tweak style #'harmonic
                ef''
            >1
            ^ \markup { III. }

        ..  figure:: ../_images/ArtificialHarmonic-cFjQuNmS3p.png

        >>> harm2 = auxjad.ArtificialHarmonic(r"<a d'>1",
        ...                                   markup='III.',
        ...                                   centre_markup=False,
        ...                                   )
        >>> abjad.show(harm2)

        ..  docs::

            <
                bf'
                \tweak style #'harmonic
                ef''
            >1
            ^ \markup { III. }

        ..  figure:: ../_images/ArtificialHarmonic-gfhbSlQPlK.png

    ..  error::

        If another markup is attached to the harmonic note, trying to set the
        :attr:`markup` property to ``None`` will raise an :exc:`Exception`:

        >>> harm = auxjad.ArtificialHarmonic(r"<a d'>1")
        >>> abjad.attach(abjad.Markup('test'), harm)
        >>> harm.markup = 'I.'
        >>> harm.markup = None
        Exception: multiple indicators attached to client.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_style",
        "_is_parenthesized",
        "_direction",
        "_centre_markup",
        "_markup",
    )

    ### INITIALISER ###

    def __init__(
        self,
        *arguments,
        multiplier: Optional[abjad.typings.DurationTyping] = None,
        tag: Optional[abjad.Tag] = None,
        style: str = "#'harmonic",
        is_parenthesized: bool = False,
        markup: Optional[str] = None,
        centre_markup: bool = True,
        direction: Union[str, abjad.enums.VerticalAlignment] = "up",
    ) -> None:
        super().__init__(*arguments, multiplier=multiplier, tag=tag)
        if len(self.written_pitches) != 2:
            raise ValueError(
                "'ArtificialHarmonic' requires exactly two 'written_pitches' for initialisation"
            )
        self.style = style
        self.is_parenthesized = is_parenthesized
        self._direction = direction
        self.centre_markup = centre_markup
        self.markup = markup

    ### PUBLIC METHODS ###

    def sounding_pitch(self) -> abjad.Pitch:
        r"""Returns the sounding pitch of the harmonic as an |abjad.Pitch|."""
        interval = abs(self.written_pitches[1] - self.written_pitches[0]).semitones
        sounding_pitch_dict = {
            1: 48,
            2: 36,
            3: 31,
            4: 28,
            5: 24,
            7: 19,
            9: 28,
            12: 12,
            16: 28,
            19: 19,
            24: 24,
            28: 28,
        }
        try:
            sounding_pitch = self.written_pitches[0] + sounding_pitch_dict[interval]
        except KeyError as err:
            raise ValueError("cannot calculate sounding pitch for given interval") from err
        return sounding_pitch

    def sounding_note(self) -> abjad.Note:
        r"""Returns the sounding note of the harmonic as an |abjad.Note|."""
        note = abjad.Note(self.sounding_pitch(), self._written_duration)
        for indicator in abjad.get.indicators(self):
            abjad.attach(indicator, note)
        return note

    ### PRIVATE METHODS ###

    def _attach_centre_markup(self) -> None:
        r"""Attaches the centre markup tweaks."""
        literal1 = abjad.LilyPondLiteral(r"\once \override TextScript.parent-alignment-X = 0")
        literal2 = abjad.LilyPondLiteral(r"\once \override TextScript.self-alignment-X = 0")
        abjad.attach(literal1, self)
        abjad.attach(literal2, self)

    def _detach_centre_markup(self) -> None:
        r"""Detaches the centre markup tweaks."""
        literal1 = abjad.LilyPondLiteral(r"\once \override TextScript.parent-alignment-X = 0")
        literal2 = abjad.LilyPondLiteral(r"\once \override TextScript.self-alignment-X = 0")
        if abjad.get.indicator(self, literal1):
            abjad.detach(literal1, self)
        if abjad.get.indicator(self, literal2):
            abjad.detach(literal2, self)

    ### PUBLIC PROPERTIES ###

    @property
    def written_pitches(self) -> abjad.pitch.segments.PitchSegment:
        r"""The written pitches of the two note heads."""
        return abjad.pitch.segments.PitchSegment(
            items=(note_head.written_pitch for note_head in self._note_heads),
            item_class=abjad.pitch.pitches.NamedPitch,
        )

    @written_pitches.setter
    def written_pitches(
        self,
        written_pitches,
    ) -> None:
        written_pitches_ = abjad.PitchSegment(written_pitches)
        if len(written_pitches_) != 2:
            raise ValueError("'ArtificialHarmonic' requires exactly two pitches")
        for index, pitch in enumerate(written_pitches_):
            self._note_heads[index].written_pitch = pitch

    @property
    def style(self) -> str:
        r"""The style of the upper note head."""
        return self._style

    @style.setter
    def style(
        self,
        style: str,
    ) -> None:
        if not isinstance(style, str):
            raise TypeError("'style' must be 'str'")
        self._style = style
        abjad.tweak(self._note_heads[1]).style = self._style

    @property
    def centre_markup(self) -> bool:
        r"""Tweaks the markup of the harmonic note head to be centred or not
        as LilyPond doesn't centralises markups above note heads by default.
        """
        return self._centre_markup

    @centre_markup.setter
    def centre_markup(
        self,
        centre_markup: bool,
    ) -> None:
        if not isinstance(centre_markup, bool):
            raise TypeError("'style' must be 'bool'")
        self._centre_markup = centre_markup

    @property
    def is_parenthesized(self) -> bool:
        r"""Whether the bottom note head is parenthesised or not."""
        return self._is_parenthesized

    @is_parenthesized.setter
    def is_parenthesized(
        self,
        is_parenthesized: bool,
    ) -> None:
        if not isinstance(is_parenthesized, bool):
            raise TypeError("'is_parenthesized' must be 'bool'")
        self._is_parenthesized = is_parenthesized
        self._note_heads[0].is_parenthesized = self._is_parenthesized
        if self._is_parenthesized:
            abjad.tweak(self._note_heads[0]).Parentheses__font_size = -3

    @property
    def markup(self) -> str:
        r"""The markup of the harmonic note head."""
        return self._markup

    @markup.setter
    def markup(
        self,
        markup: str,
    ) -> None:
        if markup is not None:
            if not isinstance(markup, str):
                raise TypeError("'markup' must be 'str'")
            self._markup = markup
            markup = abjad.Markup(
                self._markup,
                direction=self._direction,
            )
            abjad.attach(markup, self)
            if self._centre_markup:
                self._attach_centre_markup()
            else:
                self._detach_centre_markup()
        else:
            self._markup = markup
            if abjad.get.indicator(self, abjad.Markup):
                abjad.detach(abjad.Markup, self)
            self._detach_centre_markup()



from typing import Optional, Union

import abjad

from .auto_rewrite_meter import auto_rewrite_meter
from .close_container import close_container as close_container_function
from .fill_with_rests import fill_with_rests as fill_with_rests_function


def enforce_time_signature(
    container: abjad.Container,
    time_signatures: Union[
        abjad.TimeSignature,
        tuple,
        list,
    ],
    *,
    cyclic: bool = False,
    fill_with_rests: bool = True,
    close_container: bool = False,
    disable_rewrite_meter: bool = False,
    prettify_rewrite_meter: bool = True,
    boundary_depth: Optional[int] = None,
    maximum_dot_count: Optional[int] = None,
    rewrite_tuplets: bool = True,
    extract_trivial_tuplets: bool = True,
    fuse_across_groups_of_beats: bool = True,
    fuse_quadruple_meter: bool = True,
    fuse_triple_meter: bool = True,
    split_quadruple_meter: bool = True,
) -> None:
    r"""Mutates an input container (of type |abjad.Container| or child class)
    in place and has no return value; this function applies a time signature
    (or a :obj:`list` of time signatures) to the input container.

    Basic usage:
        The function mutates a container in place, applying a time signature
        to it.

        >>> staff = abjad.Staff(r"c'1 d'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
            }

        ..  figure:: ../_images/enforce_time_signature-9bf9zmnm19k.png

        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((2, 4))
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'2
                ~
                c'2
                d'2
                ~
                d'2
            }

        ..  figure:: ../_images/enforce_time_signature-kerf9uos62i.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((2, 4)),
        ... )
        >>> abjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((2, 4)),
        ... )

    Single value for second positional argument:
        The second positional argument can take either |abjad.TimeSignature|
        or a :obj:`tuple` for a single time signature (for multiple time
        signatures, use a :obj:`list` as shown further below). By default,
        rests will be appended to the end of the staff if necessary.

        >>> staff = abjad.Staff(r"c'1 d'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
            }

        ..  figure:: ../_images/enforce_time_signature-218f65bsco3.png

        >>> auxjad.mutate.enforce_time_signature(staff, (3, 4))
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                ~
                c'4
                d'2
                ~
                d'2
                r4
            }

        ..  figure:: ../_images/enforce_time_signature-u4p457k6ib7.png

    ``close_container``:
        Set the optional keyword argument ``close_container`` to ``True`` in
        order to adjust the last measure's time signature instead of filling it
        with rests.

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-tn1l53yimir.png

        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((3, 4)),
        ...     close_container=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                ~
                c'4
                d'2
                ~
                d'2
                e'4
                ~
                e'2.
                f'2.
                ~
                \time 1/4
                f'4
            }

        ..  figure:: ../_images/enforce_time_signature-1uhp08fqlpl.png

    ``fill_with_rests``:
        Alternatively, to leave the last measure as it is input (i.e. not
        filling it with rests nor adjusting the time signature), set the
        optional keyword argument ``fill_with_rests`` to ``False`` (default
        value is ``True``).

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-bit2y19hncr.png

        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((3, 4)),
        ...     fill_with_rests=False,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                ~
                c'4
                d'2
                ~
                d'2
                e'4
                ~
                e'2.
                f'2.
                ~
                f'4
            }

        ..  figure:: ../_images/enforce_time_signature-xo7fpeqsoek.png

    Multiple values for second positional argument:
        The second argument can also take a :obj:`list` of
        |abjad.TimeSignature| or :obj:`tuple`.

        >>> staff = abjad.Staff(r"c'1 d'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
            }

        ..  figure:: ../_images/enforce_time_signature-rl1csjn9osl.png

        >>> time_signatures = [abjad.TimeSignature((3, 4)),
        ...                    abjad.TimeSignature((5, 4)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                ~
                \time 5/4
                c'4
                d'1
            }

        ..  figure:: ../_images/enforce_time_signature-tqqrqi34bu.png

    Repeated time signatures:
        Consecutive identical time signatures are omitted. Also note that time
        signatures can also be represented as a :obj:`list` of :obj:`tuple`'s.

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-vn9ngz2k6cd.png

        >>> time_signatures = [(2, 4),
        ...                    (2, 4),
        ...                    (4, 4),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'2
                ~
                c'2
                \time 4/4
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-nj2c90o0pe.png

        Alternatively, use ``None`` to indicate repeated time signatures:

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-2og5ld8bkxe.png

        >>> time_signatures = [(2, 4),
        ...                    None,
        ...                    None,
        ...                    (3, 4),
        ...                    None,
        ...                    (4, 4),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'2
                ~
                c'2
                d'2
                ~
                \time 3/4
                d'2
                e'4
                ~
                e'2.
                \time 4/4
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-3s9h7p1k05x.png

    Implicit time signature:
        If the first time signature in a list is ``None``, this function
        assumes the music is written in 4/4 time signature, although it does
        not force it in the output. This is similar to LilyPond's behaviour,
        which fallbacks to a 4/4 time signature when none is present.

        >>> staff = abjad.Staff(r"c'1 d'1 e'2. f'2.")
        >>> time_signatures = [None,
        ...                    None,
        ...                    (3, 4),
        ...                    None,
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                \time 3/4
                e'2.
                f'2.
            }

        ..  figure:: ../_images/enforce_time_signature-cJe2YevEo7.png

    ``cyclic``:
        To cycle through the :obj:`list` of time signatures until the container
        is exhausted, set the optional keyword argument ``cyclic`` to ``True``.

        >>> staff = abjad.Staff(r"c'1 d'1 e'1 f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'1
                d'1
                e'1
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-vl1bwp21saq.png

        >>> time_signatures = [abjad.TimeSignature((3, 8)),
        ...                    abjad.TimeSignature((2, 8)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     time_signatures,
        ...     cyclic=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/8
                c'4.
                ~
                \time 2/8
                c'4
                ~
                \time 3/8
                c'4.
                \time 2/8
                d'4
                ~
                \time 3/8
                d'4.
                ~
                \time 2/8
                d'4
                ~
                \time 3/8
                d'8
                e'4
                ~
                \time 2/8
                e'4
                ~
                \time 3/8
                e'4.
                ~
                \time 2/8
                e'8
                f'8
                ~
                \time 3/8
                f'4.
                ~
                \time 2/8
                f'4
                ~
                \time 3/8
                f'4
                r8
            }

        ..  figure:: ../_images/enforce_time_signature-9mq64erlth6.png

    ``disable_rewrite_meter``:
        By default, this function applies the mutation
        |abjad.Meter.rewrite_meter()| to its output.

        >>> staff = abjad.Staff(r"c'1 ~ c'4 r8 d'4. e'4")
        >>> time_signatures = [abjad.TimeSignature((5, 4)),
        ...                    abjad.TimeSignature((3, 4)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/4
                c'2.
                ~
                c'2
                \time 3/4
                r8
                d'4.
                e'4
            }

        ..  figure:: ../_images/enforce_time_signature-xsjbr0vnev9.png

        To disable this, set the keyword argument ``disable_rewrite_meter`` to
        ``True``.

        >>> staff = abjad.Staff(r"c'1 ~ c'4 r8 d'4. e'4")
        >>> time_signatures = [abjad.TimeSignature((5, 4)),
        ...                    abjad.TimeSignature((3, 4)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     time_signatures,
        ...     disable_rewrite_meter=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/4
                c'1
                ~
                c'4
                \time 3/4
                r8
                d'4.
                e'4
            }

        ..  figure:: ../_images/enforce_time_signature-ezjnpwjd3xu.png

    Tuplets:
        The function handles tuplets, even if they must be split.

        >>> staff = abjad.Staff(r"\times 2/3 {c'2 d'2 e'2} f'1")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    c'2
                    d'2
                    e'2
                }
                f'1
            }

        ..  figure:: ../_images/enforce_time_signature-v4ndqpmqjk.png

        >>> time_signatures = [abjad.TimeSignature((2, 4)),
        ...                    abjad.TimeSignature((3, 4)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(staff, time_signatures)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \times 2/3
                {
                    \time 2/4
                    c'2
                    d'4
                    ~
                }
                \times 2/3
                {
                    \time 3/4
                    d'4
                    e'2
                }
                f'4
                ~
                f'2.
            }

        ..  figure:: ../_images/enforce_time_signature-5jdoukq2rkd.png

    Time signatures in the input container:
        Note that any time signatures in the input container will be ignored.

        >>> staff = abjad.Staff(r"\time 3/4 c'2. d'2. e'2. f'2.")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 3/4
                c'2.
                d'2.
                e'2.
                f'2.
            }

        ..  figure:: ../_images/enforce_time_signature-bnnz1hov5bu.png

        >>> time_signatures = [abjad.TimeSignature((5, 8)),
        ...                    abjad.TimeSignature((1, 16)),
        ...                    abjad.TimeSignature((2, 4)),
        ...                    ]
        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     time_signatures,
        ...     cyclic=True,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 5/8
                c'4.
                ~
                c'4
                ~
                \time 1/16
                c'16
                ~
                \time 2/4
                c'16
                d'4..
                ~
                \time 5/8
                d'4
                ~
                d'16
                e'16
                ~
                e'4
                ~
                \time 1/16
                e'16
                ~
                \time 2/4
                e'4.
                f'8
                ~
                \time 5/8
                f'4.
                ~
                f'4
            }

        ..  figure:: ../_images/enforce_time_signature-2l289r8sdzl.png

    Tweaking |abjad.Meter.rewrite_meter()|:
        This function uses the default logical tie splitting algorithm from
        |abjad.Meter.rewrite_meter()|.

        >>> staff = abjad.Staff(r"c'4. d'8 e'2")
        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((4, 4)),
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4.
                d'8
                e'2
            }

        ..  figure:: ../_images/enforce_time_signature-bykbobzx47.png

        Set ``boundary_depth`` to a different number to change its behaviour.

        >>> staff = abjad.Staff(r"c'4. d'8 e'2")
        >>> auxjad.mutate.enforce_time_signature(
        ...     staff,
        ...     abjad.TimeSignature((4, 4)),
        ...     boundary_depth=1,
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'4
                ~
                c'8
                d'8
                e'2
            }

        ..  figure:: ../_images/enforce_time_signature-wljhgmjh9c.png

        Other arguments available for tweaking the output of
        |abjad.Meter.rewrite_meter()| are ``maximum_dot_count`` and
        ``rewrite_tuplets``, which work exactly as the identically named
        arguments of |abjad.Meter.rewrite_meter()|.

        This function also accepts the arguments
        ``fuse_across_groups_of_beats``, ``fuse_quadruple_meter``,
        ``fuse_triple_meter``, ``extract_trivial_tuplets``, and
        ``split_quadruple_meter``, which are passed on to
        |auxjad.mutate.prettify_rewrite_meter()| (the latter can be disabled
        by setting ``prettify_rewrite_meter`` to ``False``). See the
        documentation of this function for more details on these arguments.

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/enforce_time_signature-ntl3jgbi7j.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/enforce_time_signature-y5sjtx3j0v.png

    ..  warning::

        The input container must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("first argument must be 'abjad.Container' or child class")
    if not abjad.select(container).leaves().are_contiguous_logical_voice():
        raise ValueError("first argument must be contiguous logical voice")
    if isinstance(time_signatures, list):
        time_signatures_ = time_signatures[:]
    else:
        time_signatures_ = [time_signatures]
    implicit_first_time_signature = False
    if time_signatures_[0] is None:
        time_signatures_[0] = abjad.TimeSignature((4, 4))
        implicit_first_time_signature = True
    # converting all elements to abjad.TimeSignature
    for index, time_signature in enumerate(time_signatures_):
        if time_signature is None:
            previous_ts_duration = time_signatures_[index - 1].pair
            time_signatures_[index] = abjad.TimeSignature(previous_ts_duration)
        elif not isinstance(time_signature, abjad.TimeSignature):
            time_signatures_[index] = abjad.TimeSignature(time_signature)
    partial_time_signature = None
    if time_signatures_[0].partial is not None:
        partial_time_signature = time_signatures_[0]
        time_signatures_[0] = abjad.TimeSignature(partial_time_signature.duration)
        partial_element = abjad.TimeSignature(partial_time_signature.partial)
        time_signatures_.insert(0, partial_element)
    if not isinstance(cyclic, bool):
        raise TypeError("'cyclic' must be 'bool'")
    if not isinstance(fill_with_rests, bool):
        raise TypeError("'fill_with_rests' must be 'bool'")
    if not isinstance(close_container, bool):
        raise TypeError("'close_container' must be 'bool'")
    if not isinstance(disable_rewrite_meter, bool):
        raise TypeError("'disable_rewrite_meter' must be 'bool'")
    if boundary_depth is not None:
        if not isinstance(boundary_depth, int):
            raise TypeError("'boundary_depth' must be 'int'")
    if maximum_dot_count is not None:
        if not isinstance(maximum_dot_count, int):
            raise TypeError("'maximum_dot_count' must be 'int'")
    if not isinstance(rewrite_tuplets, bool):
        raise TypeError("'rewrite_tuplets' must be 'bool'")
    if not isinstance(split_quadruple_meter, bool):
        raise TypeError("'split_quadruple_meter' must be 'bool'")
    # remove all time signatures from container
    for leaf in abjad.select(container).leaves():
        if abjad.get.indicator(leaf, abjad.TimeSignature):
            abjad.detach(abjad.TimeSignature, leaf)
    # slice container at the places where time signatures change
    durations = [time_signature.duration for time_signature in time_signatures_]
    if not cyclic:
        while sum(durations) < abjad.get.duration(container):
            durations.append(durations[-1])
    abjad.mutate.split(container[:], durations, cyclic=cyclic)
    # attach new time signatures
    previous_ts = None
    index = 0
    duration = abjad.Duration(0)
    previous_ts_duration = abjad.Duration(0)
    for leaf in abjad.select(container).leaves():
        if duration == previous_ts_duration:
            duration = abjad.Duration(0)
            previous_ts_duration = durations[index]
            if partial_time_signature is not None and index in (0, 1):
                ts = partial_time_signature
            else:
                ts = time_signatures_[index]
            if ts != previous_ts:
                abjad.attach(ts, leaf)
            previous_ts = ts
            index += 1
            if index == len(time_signatures_):
                if cyclic:
                    index = 0
                else:
                    break
        duration += abjad.get.duration(leaf)
    # filling with rests or closing container
    if close_container:
        close_container_function(container)
    elif fill_with_rests:
        fill_with_rests_function(
            container,
            disable_rewrite_meter=disable_rewrite_meter,
        )
    # removing first time signature if implicit
    if implicit_first_time_signature:
        abjad.detach(abjad.TimeSignature, abjad.select(container).leaf(0))
    # rewrite meter
    if not disable_rewrite_meter:
        measures = abjad.select(container[:]).group_by_measure()
        if cyclic:
            pattern = time_signatures_[:]
            while len(time_signatures_) < len(measures):
                time_signatures_ += pattern[:]
        else:
            while len(time_signatures_) < len(measures):
                time_signatures_.append(time_signatures_[-1])
        auto_rewrite_meter(
            container,
            meter_list=time_signatures_,
            boundary_depth=boundary_depth,
            maximum_dot_count=maximum_dot_count,
            rewrite_tuplets=rewrite_tuplets,
            prettify_rewrite_meter=prettify_rewrite_meter,
            extract_trivial_tuplets=extract_trivial_tuplets,
            fuse_across_groups_of_beats=fuse_across_groups_of_beats,
            fuse_quadruple_meter=fuse_quadruple_meter,
            fuse_triple_meter=fuse_triple_meter,
            split_quadruple_meter=split_quadruple_meter,
        )



from typing import Union

import abjad

from .. import mutate


def _make_rest_from_leaf(
    leaf: abjad.Leaf,
    *,
    ignore_dynamics: bool = False,
) -> abjad.Rest:
    r"""Creates a rest of the same duration as a leaf and passes the
    relevant leaf indicators to the rest (such as time signature or bar
    lines, but not slurs or articulations)
    """
    rest = abjad.Rest(leaf.written_duration)
    for indicator in abjad.get.indicators(leaf):
        if isinstance(
            indicator,
            (
                abjad.BarLine,
                abjad.Clef,
                abjad.Dynamic,
                abjad.KeySignature,
                abjad.LilyPondLiteral,
                abjad.MetronomeMark,
                abjad.RehearsalMark,
                abjad.Repeat,
                abjad.StaffChange,
                abjad.StartHairpin,
                abjad.StartSlur,
                abjad.StopHairpin,
                abjad.StopSlur,
                abjad.TimeSignature,
            ),
        ):
            abjad.attach(indicator, rest)
    return rest


def staff_splitter(
    staff: Union[abjad.Staff, abjad.Selection],
    *,
    threshold: Union[
        int,
        float,
        str,
        abjad.Pitch,
    ] = abjad.NamedPitch("c'"),
    upper_clef: Union[abjad.Clef, str] = abjad.Clef("treble"),
    lower_clef: Union[abjad.Clef, str] = abjad.Clef("bass"),
    add_clefs: bool = True,
    dynamics_only_on_upper_staff: bool = False,
    reposition_dynamics: bool = True,
    reposition_slurs: bool = True,
    use_multimeasure_rests: bool = True,
    rewrite_meter: bool = True,
) -> tuple:
    r"""Takes an |abjad.Staff| or |abjad.Selection| and splits it into two
    staves using a reference pitch as threshold. Returns a tuple of
    |abjad.Staff|'s.

    Basic usage:
        By default, this function splits notes using ``C4`` as the pitch
        threshold (any pitches equal to or higher than the threhold will be
        added to the upper staff, and any pitches lower than it will be added
        to the lower staff). It also automatically adds clefs to both staves.
        This function returns a tuple of two |abjad.Staff|'s, which can be
        input directly into an |abjad.Score|:

        >>> staff = abjad.Staff(r"a4 b4 c'4 d'4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                a4
                b4
                c'4
                d'4
            }

        ..  figure:: ../_images/staff_splitter-h3b7QkOtTm.png

        >>> staves = auxjad.staff_splitter(staff)
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    r2
                    c'4
                    d'4
                }
                \new Staff
                {
                    \clef "bass"
                    a4
                    b4
                    r2
                }
            >>

        ..  figure:: ../_images/staff_splitter-WQoeWAkkmV.png

    Chords:
        Chords are split according to the threhold pitch:

        >>> staff = abjad.Staff(r"<g b>4 <a c'>4 <b d' f'>4 <a f c' e' g'>4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                <g b>4
                <a c'>4
                <b d' f'>4
                <a f c' e' g'>4
            }

        ..  figure:: ../_images/staff_splitter-ICrujezDfk.png

        >>> staves = auxjad.staff_splitter(staff)
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    r4
                    c'4
                    <d' f'>4
                    <c' e' g'>4
                }
                \new Staff
                {
                    \clef "bass"
                    <g b>4
                    a4
                    b4
                    <f a>4
                }
            >>

        ..  figure:: ../_images/staff_splitter-KASqlXv1Fh.png

    ``threshold``:
        Set the keword argument ``threshold`` to a pitch to specify the
        threhold pitch. Notes lower than this pitch will be added to the lower
        staff whereas notes higher than or equal to this pitch will be added to
        the upper one. It can take an |abjad.NamedPitch|, an
        |abjad.NumberedPitch|, or a :obj:`str`, :obj:`int`, or :obj:`float`.

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 <d' f' a'>4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                <d' f' a'>4
            }

        ..  figure:: ../_images/staff_splitter-nbVARRMPpG.png

        >>> staves = auxjad.staff_splitter(staff, threshold="e'")
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    r2
                    e'4
                    <f' a'>4
                }
                \new Staff
                {
                    \clef "bass"
                    c'4
                    d'4
                    r4
                    d'4
                }
            >>

        ..  figure:: ../_images/staff_splitter-hwD9HkFusx.png

    ``lower_clef`` and ``upper_clef``:
        By default, the clefs of the lower and upper staves are set to bass and
        treble clefs, respectively. Use the keyword arguments ``lower_clef``
        and ``upper_clef`` to change this. They can take either an |abjad.Clef|
        or an :obj:`str`.

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 <d' f' a'>4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                <d' f' a'>4
            }

        ..  figure:: ../_images/staff_splitter-eJBqwmhPsj.png

        >>> staves = auxjad.staff_splitter(staff,
        ...                                threshold="e'",
        ...                                lower_clef='treble',
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    r2
                    e'4
                    <f' a'>4
                }
                \new Staff
                {
                    \clef "treble"
                    c'4
                    d'4
                    r4
                    d'4
                }
            >>

        ..  figure:: ../_images/staff_splitter-87mg2UBDPn.png

        >>> staff = abjad.Staff(r"e4 f4 g4 <f a c'>4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                e4
                f4
                g4
                <f a c'>4
            }

        ..  figure:: ../_images/staff_splitter-Nu9dNMM9gy.png

        >>> staves = auxjad.staff_splitter(staff,
        ...                                threshold='g',
        ...                                upper_clef='bass',
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "bass"
                    r2
                    g4
                    <a c'>4
                }
                \new Staff
                {
                    \clef "bass"
                    e4
                    f4
                    r4
                    f4
                }
            >>

        ..  figure:: ../_images/staff_splitter-68Qhb2RuTj.png

    ``add_clefs``:
        To not add clefs to the output staves, set ``add_clefs`` to ``False``:

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 <d' f' a'>4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'4
                d'4
                e'4
                <d' f' a'>4
            }

        ..  figure:: ../_images/staff_splitter-uoSjmalLzc.png

        >>> staves = auxjad.staff_splitter(staff,
        ...                                threshold="e'",
        ...                                add_clefs=False,
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    r2
                    e'4
                    <f' a'>4
                }
                \new Staff
                {
                    c'4
                    d'4
                    r4
                    d'4
                }
            >>

        ..  figure:: ../_images/staff_splitter-roBNy3fRus.png

        ..  note::

            Note how there are no clefs added to the LilyPond output. Note that
            in the image above that LilyPond automatically fallbacks to treble
            clefs when no clef is present in the input code.

            >>> staff = abjad.Staff(r"c'4 d'4 e'4 <d' f' a'>4")
            >>> staves = auxjad.staff_splitter(staff,
            ...                                threshold="e'",
            ...                                add_clefs=False,
            ...                                )
            >>> string = abjad.lilypond(score)
            >>> print(string)
            \new Score
            <<
                \new Staff
                {
                    r2
                    e'4
                    <f' a'>4
                }
                \new Staff
                {
                    c'4
                    d'4
                    r4
                    d'4
                }
            >>

    ``use_multimeasure_rests``:
        By default, rests are converted to multi-measure rests.

        >>> staff = abjad.Staff(
        ...     r"\time 2/4 c'2 \times 2/3 {<g b d'>2 <e' f'>4}"
        ...     r"\times 2/3 {a2 <g b>4}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                c'2
                \times 2/3
                {
                    <g b d'>2
                    <e' f'>4
                }
                \times 2/3
                {
                    a2
                    <g b>4
                }
            }

        ..  figure:: ../_images/staff_splitter-QkVo9lbpvB.png

        >>> staves = auxjad.staff_splitter(staff)
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 2/4
                    \clef "treble"
                    c'2
                    \times 2/3
                    {
                        d'2
                        <e' f'>4
                    }
                    R1 * 1/2
                }
                \new Staff
                {
                    \time 2/4
                    \clef "bass"
                    R1 * 1/2
                    \times 2/3
                    {
                        <g b>2
                        r4
                    }
                    \times 2/3
                    {
                        a2
                        <g b>4
                    }
                }
            >>

        ..  figure:: ../_images/staff_splitter-gLP4jNpRXE.png

        Set ``use_multimeasure_rests`` to ``False`` to disable this behaviour.

        >>> staves = auxjad.staff_splitter(staff,
        ...                                use_multimeasure_rests=False,
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 2/4
                    \clef "treble"
                    c'2
                    \times 2/3
                    {
                        d'2
                        <e' f'>4
                    }
                    r2
                }
                \new Staff
                {
                    \time 2/4
                    \clef "bass"
                    r2
                    \times 2/3
                    {
                        <g b>2
                        r4
                    }
                    \times 2/3
                    {
                        a2
                        <g b>4
                    }
                }
            >>

        ..  figure:: ../_images/staff_splitter-s2Q7w1aVAC.png

    ``rewrite_meter``:
        By default, the |abjad.Meter.rewrite_meter()| mutation is applied to
        the output. Set ``rewrite_meter`` to ``False`` to disable it:

        >>> staff = abjad.Staff(r"a4 b4 c'4 d'4")
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                a4
                b4
                c'4
                d'4
            }

        ..  figure:: ../_images/staff_splitter-JyRf5fsouP.png

        >>> staves = auxjad.staff_splitter(staff,
        ...                                rewrite_meter=False,
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    r4
                    r4
                    c'4
                    d'4
                }
                \new Staff
                {
                    \clef "bass"
                    a4
                    b4
                    r4
                    r4
                }
            >>

        ..  figure:: ../_images/staff_splitter-hinmixCe4v.png

    ``reposition_dynamics``:
        By default, dynamics are split among the staves and then are mutated
        using :func:`auxjad.mutate.reposition_dynamics()`.

        >>> staff = abjad.Staff(
        ...     r"c'2\p <b d'>2\ff \times 2/3 {<g b d'>2\f <e' f'>1\mf}"
        ...     r"\times 2/3 {a2\pp <g b>1\mp}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                \p
                <b d'>2
                \ff
                \times 2/3
                {
                    <g b d'>2
                    \f
                    <e' f'>1
                    \mf
                }
                \times 2/3
                {
                    a2
                    \pp
                    <g b>1
                    \mp
                }
            }

        ..  figure:: ../_images/staff_splitter-I42iFOOMEm.png

        >>> staves = auxjad.staff_splitter(staff)
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    c'2
                    \p
                    d'2
                    \ff
                    \times 2/3
                    {
                        d'2
                        \f
                        <e' f'>1
                        \mf
                    }
                    R1
                }
                \new Staff
                {
                    \clef "bass"
                    r2
                    b2
                    \ff
                    \times 2/3
                    {
                        <g b>2
                        \f
                        r1
                    }
                    \times 2/3
                    {
                        a2
                        \pp
                        <g b>1
                        \mp
                    }
                }
            >>

        ..  figure:: ../_images/staff_splitter-SYTT14sWJo.png

        Set ``reposition_dynamics`` to ``False`` to disable this behaviour:

        >>> staff = abjad.Staff(
        ...     r"c'2\p <b d'>2\ff \times 2/3 {<g b d'>2\f <e' f'>1\mf}"
        ...     r"\times 2/3 {a2\pp <g b>1\mp}"
        ... )
        >>> staves = auxjad.staff_splitter(staff,
        ...                                reposition_dynamics=False,
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    c'2
                    \p
                    d'2
                    \ff
                    \times 2/3
                    {
                        d'2
                        \f
                        <e' f'>1
                        \mf
                    }
                    R1
                    \mp
                }
                \new Staff
                {
                    \clef "bass"
                    r2
                    \p
                    b2
                    \ff
                    \times 2/3
                    {
                        <g b>2
                        \f
                        r1
                        \mf
                    }
                    \times 2/3
                    {
                        a2
                        \pp
                        <g b>1
                        \mp
                    }
                }
            >>

        ..  figure:: ../_images/staff_splitter-sXrC86fy1J.png

        ..  note::

            It is important to note that dynamics can easily become problematic
            in the output when dealing with complex input staves. When
            possible, it is always best to handle containers without dynamics
            or those with very simple dynamics.

    ``dynamics_only_on_upper_staff``:
        Dynamics are distributed to both staves by default. To keep the
        dynamics only to the upper staff, set ``dynamics_only_on_upper_staff``
        to ``True``

        >>> staff = abjad.Staff(
        ...     r"c'2\p <b d'>2\ff \times 2/3 {<g b d'>2\f <e' f'>1\mf}"
        ...     r"\times 2/3 {a2\pp <g b>1\mp}"
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                c'2
                \p
                <b d'>2
                \ff
                \times 2/3
                {
                    <g b d'>2
                    \f
                    <e' f'>1
                    \mf
                }
                \times 2/3
                {
                    a2
                    \pp
                    <g b>1
                    \mp
                }
            }

        ..  figure:: ../_images/staff_splitter-f6E9w8Lpn7.png

        >>> staves = auxjad.staff_splitter(staff,
        ...                                dynamics_only_on_upper_staff=True,
        ...                                )
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \clef "treble"
                    c'2
                    \p
                    d'2
                    \ff
                    \times 2/3
                    {
                        d'2
                        \f
                        <e' f'>1
                        \mf
                    }
                    R1
                }
                \new Staff
                {
                    \clef "bass"
                    r2
                    b2
                    \times 2/3
                    {
                        <g b>2
                        r1
                    }
                    \times 2/3
                    {
                        a2
                        <g b>1
                    }
                }
            >>

        ..  figure:: ../_images/staff_splitter-bVcVinE5cG.png

        ..  note::

            Similarly to the previous note about dynamics, it is important to
            note that, once again, dynamics can easily become problematic in
            the output. If there are rests or multi-measure rests in the upper
            staff, the dynamics can get lost. The best approach is to add them
            after the split.

    Slurs and articulations:
        This function will automatically handle slurs and articulations:

        >>> staff = abjad.Staff(
        ...     r"\time 2/4 a8( b c' d') \times 2/3 {<g b d'>2 <e' f'>4}"
        ...     r"\time 3/4 <d a c' g'>4--  r8 <f a bf>4."
        ... )
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 2/4
                a8
                (
                b8
                c'8
                d'8
                )
                \times 2/3
                {
                    <g b d'>2
                    <e' f'>4
                }
                \time 3/4
                <d a c' g'>4
                - \tenuto
                r8
                <f a bf>4.
            }

        ..  figure:: ../_images/staff_splitter-SZjT1IJCsn.png

        >>> staves = auxjad.staff_splitter(staff)
        >>> score = abjad.Score(staves)
        >>> abjad.show(score)

        ..  docs::

            \new Score
            <<
                \new Staff
                {
                    \time 2/4
                    \clef "treble"
                    r4
                    c'8
                    (
                    d'8
                    )
                    \times 2/3
                    {
                        d'2
                        <e' f'>4
                    }
                    \time 3/4
                    <c' g'>4
                    - \tenuto
                    r2
                }
                \new Staff
                {
                    \time 2/4
                    \clef "bass"
                    a8
                    (
                    b8
                    )
                    r4
                    \times 2/3
                    {
                        <g b>2
                        r4
                    }
                    \time 3/4
                    <d a>4
                    - \tenuto
                    r8
                    <f a bf>4.
                }
            >>

        ..  figure:: ../_images/staff_splitter-RbPPaLUzjO.png
    """
    if not isinstance(staff, (abjad.Staff, abjad.Selection)):
        raise TypeError("'staff' must be 'abjad.Staff' or 'abjad.Selection'")
    if isinstance(threshold, str):
        threshold = abjad.NamedPitch(threshold)
    elif isinstance(threshold, (int, float)):
        threshold = abjad.NumberedPitch(threshold)
    elif not isinstance(threshold, abjad.Pitch):
        raise TypeError(
            "'threshold' must be 'abjad.Pitch' (or child class), 'str', 'int', or 'float'"
        )
    if isinstance(upper_clef, str):
        upper_clef = abjad.Clef(upper_clef)
    elif not isinstance(upper_clef, abjad.Clef):
        raise TypeError("'upper_clef' must be 'abjad.Clef' or 'str'")
    if isinstance(lower_clef, str):
        lower_clef = abjad.Clef(lower_clef)
    elif not isinstance(lower_clef, abjad.Clef):
        raise TypeError("'lower_clef' must be 'abjad.Clef' or 'str'")
    if not isinstance(add_clefs, bool):
        raise TypeError("'add_clefs' must be 'bool'")
    if not isinstance(dynamics_only_on_upper_staff, bool):
        raise TypeError("'dynamics_only_on_upper_staff' must be 'bool'")
    if not isinstance(reposition_dynamics, bool):
        raise TypeError("'reposition_dynamics' must be 'bool'")
    if not isinstance(reposition_slurs, bool):
        raise TypeError("'reposition_slurs' must be 'bool'")
    if not isinstance(use_multimeasure_rests, bool):
        raise TypeError("'use_multimeasure_rests' must be 'bool'")
    if not isinstance(rewrite_meter, bool):
        raise TypeError("'rewrite_meter' must be 'bool'")

    upper_staff = abjad.mutate.copy(staff)
    lower_staff = abjad.mutate.copy(staff)
    if isinstance(staff, abjad.Selection):
        upper_staff = abjad.Staff(upper_staff)
        lower_staff = abjad.Staff(lower_staff)

    for leaf in abjad.select(upper_staff).leaves():
        if isinstance(leaf, abjad.Note) and leaf.written_pitch < threshold:
            rest = _make_rest_from_leaf(leaf)
            abjad.mutate.replace(leaf, rest)
        elif isinstance(leaf, abjad.Chord):
            if all([pitch < threshold for pitch in leaf.written_pitches]):
                rest = _make_rest_from_leaf(leaf)
                abjad.mutate.replace(leaf, rest)
            else:
                new_pitches = []
                for pitch in leaf.written_pitches:
                    if pitch >= threshold:
                        new_pitches.append(pitch)
                if len(new_pitches) > 1:
                    leaf.written_pitches = new_pitches
                else:
                    new_leaf = abjad.Note(
                        new_pitches[0],
                        leaf.written_duration,
                    )
                    for indicator in abjad.get.indicators(leaf):
                        abjad.attach(indicator, new_leaf)
                    abjad.mutate.replace(leaf, new_leaf)

    for leaf in abjad.select(lower_staff).leaves():
        if isinstance(leaf, abjad.Note):
            if leaf.written_pitch >= threshold:
                rest = _make_rest_from_leaf(leaf)
                abjad.mutate.replace(leaf, rest)
        if isinstance(leaf, abjad.Chord):
            if all([pitch >= threshold for pitch in leaf.written_pitches]):
                rest = _make_rest_from_leaf(leaf)
                abjad.mutate.replace(leaf, rest)
            else:
                new_pitches = []
                for pitch in leaf.written_pitches:
                    if pitch < threshold:
                        new_pitches.append(pitch)
                if len(new_pitches) > 1:
                    leaf.written_pitches = new_pitches
                else:
                    new_leaf = abjad.Note(
                        new_pitches[0],
                        leaf.written_duration,
                    )
                    for indicator in abjad.get.indicators(leaf):
                        abjad.attach(indicator, new_leaf)
                    abjad.mutate.replace(leaf, new_leaf)

    if add_clefs:
        abjad.attach(upper_clef, abjad.select(upper_staff).leaf(0))
        abjad.attach(lower_clef, abjad.select(lower_staff).leaf(0))

    if dynamics_only_on_upper_staff:
        for leaf in abjad.select(lower_staff).leaves():
            for indicator in abjad.get.indicators(leaf):
                if isinstance(
                    indicator,
                    (
                        abjad.Dynamic,
                        abjad.StartHairpin,
                        abjad.StopHairpin,
                    ),
                ):
                    abjad.detach(indicator, leaf)

    if reposition_dynamics:
        mutate.reposition_dynamics(upper_staff[:])
        if not dynamics_only_on_upper_staff:
            mutate.reposition_dynamics(lower_staff[:])

    if reposition_slurs:
        mutate.reposition_slurs(upper_staff[:])
        mutate.reposition_slurs(lower_staff[:])

    if use_multimeasure_rests:
        mutate.rests_to_multimeasure_rest(upper_staff[:])
        mutate.rests_to_multimeasure_rest(lower_staff[:])

    if rewrite_meter:
        mutate.auto_rewrite_meter(upper_staff)
        mutate.auto_rewrite_meter(lower_staff)

    return upper_staff, lower_staff


### EXTENSION FUNCTIONS ###

abjad.staff_splitter = staff_splitter



from typing import Optional

import abjad

from .. import get
from .prettify_rewrite_meter import (
    prettify_rewrite_meter as prettify_rewrite_meter_function,
)


def fill_with_rests(
    container: abjad.Container,
    *,
    disable_rewrite_meter: bool = False,
    prettify_rewrite_meter: bool = True,
    boundary_depth: Optional[int] = None,
    maximum_dot_count: Optional[int] = None,
    rewrite_tuplets: bool = True,
    extract_trivial_tuplets: bool = True,
    fuse_across_groups_of_beats: bool = True,
    fuse_quadruple_meter: bool = True,
    fuse_triple_meter: bool = True,
    split_quadruple_meter: bool = True,
) -> None:
    r"""Mutates an input container (of type |abjad.Container| or child class)
    in place and has no return value; this function fills a container with
    rests in order to make it full.

    Basic usage:
        Returns the missing duration of the last measure of any container or
        child class. If no time signature is encountered, it uses LilyPond's
        convention and considers the container as in 4/4.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.mutate.fill_with_rests(container1)
        >>> auxjad.mutate.fill_with_rests(container2)
        >>> auxjad.mutate.fill_with_rests(container3)
        >>> auxjad.mutate.fill_with_rests(container4)
        >>> abjad.show(container1)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/fill_with_rests-up297scg6t.png

        >>> abjad.show(container2)

        ..  docs::

            {
                c'4
                d'4
                e'4
                r4
            }

        ..  figure:: ../_images/fill_with_rests-azrrw0z0buw.png

        >>> abjad.show(container3)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
                c'4
                r2.
            }

        ..  figure:: ../_images/fill_with_rests-jtyw5ikc0k.png

        >>> abjad.show(container4)

        ..  docs::

            {
                c'4
                d'4
                e'4
                f'4
                c'4
                d'4
                e'4
                f'4
            }

        ..  figure:: ../_images/fill_with_rests-xjkm2vzjfpr.png

    ..  note::

        Auxjad automatically adds this function as an extension function to
        |abjad.mutate|. It can thus be used from either |auxjad.mutate|_ or
        |abjad.mutate| namespaces. Therefore, the two lines below are
        equivalent:

        >>> auxjad.mutate.fill_with_rests(staff)
        >>> abjad.mutate.fill_with_rests(staff)

    Time signature changes:
        Handles any time signatures as well as changes of time signature.

        >>> staff1 = abjad.Staff(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
        >>> staff2 = abjad.Staff(r"\time 3/4 a2. \time 2/4 c'4")
        >>> staff3 = abjad.Staff(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> auxjad.mutate.fill_with_rests(staff1)
        >>> auxjad.mutate.fill_with_rests(staff2)
        >>> auxjad.mutate.fill_with_rests(staff3)
        >>> abjad.show(staff1)

        ..  docs::

            {
                \time 4/4
                c'4
                d'4
                e'4
                f'4
                g'4
                r2.
            }

        ..  figure:: ../_images/fill_with_rests-7zydps2jsb.png

        >>> abjad.show(staff2)

        ..  docs::

            {
                \time 3/4
                a2.
                \time 2/4
                c'4
                r4
            }

        ..  figure:: ../_images/fill_with_rests-1lphcrl5pnr.png

        >>> abjad.show(staff3)

        ..  docs::

            {
                \time 5/4
                g1
                ~
                g4
                \time 4/4
                af'2
                r2
            }

        ..  figure:: ../_images/fill_with_rests-e00jvx986r.png

    ..  note::

        When using |abjad.Container|'s, all time signatures in the output will
        be commented out with ``%%%.`` This is because Abjad only applies time
        signatures to containers that belong to a |abjad.Staff|. The present
        function works with either |abjad.Container| and |abjad.Staff|.

        >>> container = abjad.Container(r"\time 3/4 c'4 d'4 e'4")
        >>> abjad.show(container)

        ..  docs::

            {
                %%% \time 3/4 %%%
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/fill_with_rests-b0qflg50qfn.png

        >>> staff = abjad.Staff([container])
        >>> abjad.show(container)

        ..  docs::

            {
                \time 3/4
                c'4
                d'4
                e'4
            }

        ..  figure:: ../_images/fill_with_rests-qtaswjiecg.png

    Partial time signatures:
        Correctly handles partial time signatures.

        >>> staff = abjad.Staff(r"c'4 d'4 e'4 f'4 g'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, staff[0])
        >>> auxjad.mutate.fill_with_rests(staff)
        >>> abjad.show(staff)

        ..  docs::

            {
                \partial 4
                \time 3/4
                c'4
                d'4
                e'4
                f'4
                g'4
                r2
            }

        ..  figure:: ../_images/fill_with_rests-9smva9ajdi.png

    ``disable_rewrite_meter``:
        By default, this class applies the |abjad.Meter.rewrite_meter()|
        mutation to the last measure when rests are added.

        >>> staff = abjad.Staff(r"\time 4/4 c'8 d'4 e'4")
        >>> auxjad.mutate.fill_with_rests(staff)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'4
                e'8
                ~
                e'8
                r4.
            }

        ..  figure:: ../_images/fill_with_rests-n83nmnfh92c.png

        Call this function with the optional keyword argument
        ``disable_rewrite_meter`` set to ``True`` in order to disable this
        behaviour.

        >>> staff = abjad.Staff(r"\time 4/4 c'8 d'4 e'4")
        >>> auxjad.mutate.fill_with_rests(staff, disable_rewrite_meter=True)
        >>> abjad.show(staff)

        ..  docs::

            \new Staff
            {
                \time 4/4
                c'8
                d'4
                e'4
                r4.
            }

        ..  figure:: ../_images/fill_with_rests-9rg2i4n1vhr.png

    ..  note::

        This function also accepts the arguments ``boundary_depth``,
        ``maximum_dot_count``, and ``rewrite_tuplets``, which are passed on to
        |abjad.Meter.rewrite_meter()|, and ``fuse_across_groups_of_beats``,
        ``fuse_quadruple_meter``, ``fuse_triple_meter``,
        ``extract_trivial_tuplets``,  and ``split_quadruple_meter``, which are
        passed on to |auxjad.mutate.prettify_rewrite_meter()| (the latter can
        be disabled by setting ``prettify_rewrite_meter`` to ``False``). See
        the documentation of those functions for more details on these
        arguments.

    ..  error::

        If a container is malformed, i.e. it has an underfilled measure before
        a time signature change, the function raises a :exc:`ValueError`
        exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'4")
        >>> auxjad.mutate.fill_with_rests(container)
        ValueError: 'container' is malformed, with an underfull measure
        preceding a time signature change

    ..  warning::

        The input container must be a contiguous logical voice. When dealing
        with a container with multiple subcontainers (e.g. a score containing
        multiple staves), the best approach is to cycle through these
        subcontainers, applying this function to them individually.
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("argument must be 'abjad.Container' or child class")
    if not abjad.select(container).leaves().are_contiguous_logical_voice():
        raise ValueError("argument must be contiguous logical voice")
    try:
        if not get.selection_is_full(container[:]):
            underfull_rests = abjad.LeafMaker()(
                None,
                get.underfull_duration(container[:]),
            )
            container.extend(underfull_rests)
        else:
            return
    except ValueError as err:
        raise ValueError(
            "'container' is malformed, with an underfull measure "
            "preceding a time signature change"
        ) from err
    if not disable_rewrite_meter:
        time_signatures = get.time_signature_list(
            container,
            do_not_use_none=True,
        )
        measures = abjad.select(container[:]).group_by_measure()
        abjad.Meter.rewrite_meter(
            measures[-1],
            time_signatures[-1],
            boundary_depth=boundary_depth,
            maximum_dot_count=maximum_dot_count,
            rewrite_tuplets=rewrite_tuplets,
        )
        if prettify_rewrite_meter:
            measures = abjad.select(container[:]).group_by_measure()
            prettify_rewrite_meter_function(
                measures[-1],
                time_signatures[-1],
                extract_trivial_tuplets=extract_trivial_tuplets,
                fuse_across_groups_of_beats=fuse_across_groups_of_beats,
                fuse_quadruple_meter=fuse_quadruple_meter,
                fuse_triple_meter=fuse_triple_meter,
                split_quadruple_meter=split_quadruple_meter,
            )
